import os
import argparse

def convert(fpath):

    assert fpath.endswith('h264'), 'Source file must have .h264 file name extension.'

    target_fpath = fpath[:-5] + '.avi'

    # command = 'ffmpeg -i ' + fpath + \
    #           ' -pix_fmt yuv420p ' + \
    #           '-vf scale=w=-1:h=-1:in_range=pc:out_range=tv ' + \
    #           '-c:v libx264 ' + target_fpath

    command = 'ffmpeg -i ' + fpath + \
              ' -pix_fmt yuv420p ' + \
              '-vf eq=gamma=1.5:saturation=1.5 ' + \
              '-c:v libx264 ' + target_fpath

    os.system(command)

# Running the following command on the output .avi file yields good luminance quality
# ffmpeg -i original.avi -vf eq=gamma=1.5:saturation=1.5 -c:a copy  outfile.avi


if __name__ == '__main__':
    # Input argument handling and help info
    parser = argparse.ArgumentParser(description='Convert h264 to mjpeg, conserving full color range.')
    parser.add_argument('source_path', type=str, nargs=1, 
                        help='Path to H264 encoded video.')
    args = parser.parse_args()
    convert(args.source_path[0])



### This script combines position data from multiple cameras.
### It also corrects frame time offset errors in PosLog.csv files
### It also removes bad position data lines

### Use as follows:
### import CombineTrackingData as combPos
### combPos.combdata(Path-To-Recording-Folder)

### By Sander Tanni, May 2017, UCL

import os
from multiprocessing import Process
from itertools import combinations
import argparse

import numpy as np
from scipy.spatial.distance import euclidean
import h5py
from tqdm import tqdm


def combineCamerasData(cameraPos, lastCombPos, cameraIDs, CameraSettings, arena_size):
    # This outputs position data based on which camera is closest to tracking target.

    # cameraPos - list of numpy vecors with 4 elements (x1,y1,x2,y2) for each camera
    # lastCombPos - Last known output from this function. If None, the function will attempt to locate the animal.
    # cameraIDs - list of of CameraSettings.keys() in corresponding order to cameraPos and lastCombPos
    # CameraSettings - settings dictionary created by CameraSettings.CameraSettingsApp
    # arena_size - 2 element numpy array with arena height and width.

    # Output - numpy vector (x1,y1,x2,y2) with data from closest camera

    # Animal detection finding method in case lastCombPos=None
    #   simultaneously from at least 2 cameras with smaller separation than half of CameraSettings['camera_transfer_radius']
    #   If successful, closest mean coordinate is set as output
    #   If unsuccessful, output is None

    N_RPis = len(cameraPos)
    cameraPos = np.array(cameraPos, dtype=np.float32)
    camera_locations = []
    for cameraID in cameraIDs:
        camera_locations.append(CameraSettings['CameraSpecific'][cameraID]['location_xy'])
    camera_locations = np.array(camera_locations, dtype=np.float32)

    # Only work with camera data from inside the enviornment
    # Find bad pos data lines
    idxBad = np.zeros(cameraPos.shape[0], dtype=bool)
    # Points beyond arena size
    x_too_big = cameraPos[:,0] > arena_size[0] + 20
    y_too_big = cameraPos[:,1] > arena_size[1] + 20
    idxBad = np.logical_or(idxBad, np.logical_or(x_too_big, y_too_big))
    x_too_small = cameraPos[:,0] < -20
    y_too_small = cameraPos[:,1] < -20
    idxBad = np.logical_or(idxBad, np.logical_or(x_too_small, y_too_small))
    # Only keep camera data from within the environment
    N_RPis = np.sum(np.logical_not(idxBad))
    # Only continue if at least one RPi data remains
    if N_RPis > 0:
        cameraPos = cameraPos[np.logical_not(idxBad),:]
        camera_locations = camera_locations[np.logical_not(idxBad),:]
        if np.any(lastCombPos):
            # Check which cameras provide data close enough to lastCombPos
            RPi_correct = []
            for nRPi in range(N_RPis):
                lastCombPos_distance = euclidean(cameraPos[nRPi, :2], lastCombPos[:2])
                RPi_correct.append(lastCombPos_distance < CameraSettings['General']['camera_transfer_radius'])
            RPi_correct = np.array(RPi_correct, dtype=bool)
            # If none were found to be withing search radius, set output to None
            if not np.any(RPi_correct):
                combPos = None
            else:
                # Use the reading from closest camera to target mean location that detects correct location
                if np.sum(RPi_correct) > 1:
                    # Only use correct cameras
                    N_RPis = np.sum(RPi_correct)
                    cameraPos = cameraPos[RPi_correct, :]
                    camera_locations = camera_locations[RPi_correct, :]
                    meanPos = np.mean(cameraPos[:, :2], axis=0)
                    # Find mean position distance from all cameras
                    cam_distances = []
                    for nRPi in range(N_RPis):
                        camera_loc = camera_locations[nRPi, :]
                        cam_distances.append(euclidean(camera_loc, meanPos))
                    # Find closest distance camera and output its location coordinates
                    closest_camera = np.argmin(np.array(cam_distances))
                    combPos = cameraPos[closest_camera, :]
                else:
                    # If target only detected close enough to lastCombPos in a single camera, use it as output
                    combPos = cameraPos[np.where(RPi_correct)[0][0], :]
        else:
            # If no lastCombPos provided, check if position can be verified from more than one camera
            #### NOTE! This solution breaks down if more than two cameras incorrectly identify the same object
            ####       as the brightes spot, instead of the target LED.
            cameraPairs = []
            pairDistances = []
            for c in combinations(range(N_RPis), 2):
                pairDistances.append(euclidean(cameraPos[c[0], :2], cameraPos[c[1], :2]))
                cameraPairs.append(np.array(c))
            cameraPairs = np.array(cameraPairs)
            cameraPairs_Match = np.array(pairDistances) < (CameraSettings['General']['camera_transfer_radius'] / 2.0)
            # If position can not be verified from multiple cameras, set output to none
            if not np.any(cameraPairs_Match):
                combPos = None
            else:
                # Otherwise, set output to mean of two cameras with best matching detected locations
                pairToUse = np.argmin(pairDistances)
                camerasToUse = np.array(cameraPairs[pairToUse, :])
                combPos = np.mean(cameraPos[camerasToUse, :2], axis=0)
                # Add NaN values for second LED
                combPos = np.append(combPos, np.empty(2) * np.nan)
    else:
        combPos = None

    return combPos

def estimate_OpenEphys_timestamps_for_tracking_data(OE_GC_times, RPi_GC_times, RPi_frame_times, verbose=False):
    '''
    Estimates Open Ephys timestamps for each tracking data datapoint.
    OE_GC_times - Global Clock timestamps in Open Ephys time
    RPi_GC_times - Global Clock timestamps in RPi time
    RPi_frame_times - Frame timestamps in RPi time
    '''
    GlobalClock_TTL_Channel = 1
    RPiTime2Sec = 10 ** 6 # This values is used to convert RPi times to seconds
    # Crop data if more timestamps recorded on either system.
    if OE_GC_times.size > RPi_GC_times.size:
        OE_GC_times = OE_GC_times[:RPi_GC_times.size]
        print('[ Warning ] OpenEphys recorded more GlobalClock TTL pulses than RPi.\n' + 
              'Dumping extra timestamps from the end.')
    elif OE_GC_times.size < RPi_GC_times.size:
        RPi_GC_times = RPi_GC_times[:OE_GC_times.size]
        print('[ Warning ] RPi recorded more GlobalClock TTL pulses than Open Ephys.\n' + 
              'Dumping extra timestamps from the end.')
    # Find closest RPi GlobalClock timestamp to each RPi frame timestamp
    if verbose:
        print('Estimating camera timestamps in OpenEPhys time')
    RPtimes_idx = []
    for RPFtime in (tqdm(RPi_frame_times) if verbose else RPi_frame_times):
        RPtimes_idx.append(np.argmin(np.abs(RPi_GC_times - RPFtime)))
    # Compute difference from the GlobalClock timestamps for each frame
    RPtimes_full = RPi_GC_times[RPtimes_idx]
    RPF_RP_times_delta = RPi_frame_times - RPtimes_full
    # Convert this difference to seconds as Open Ephys timestamps
    RPF_RP_times_delta_in_seconds = RPF_RP_times_delta / float(RPiTime2Sec)
    # Use frame and GlobalClock timestamp diffence to estimate OpenEphys timepoints
    OEtimes_full = OE_GC_times[RPtimes_idx]
    RPi_frame_in_OE_times = OEtimes_full + RPF_RP_times_delta_in_seconds

    return RPi_frame_in_OE_times

def remove_tracking_data_outside_boundaries(posdata, arena_size, max_error=20):
    NotNaN = np.where(np.logical_not(np.isnan(posdata[:,1])))[0]
    idxBad = np.zeros(NotNaN.size, dtype=bool)
    x_too_big = posdata[NotNaN,1] > arena_size[0] + max_error
    y_too_big = posdata[NotNaN,2] > arena_size[1] + max_error
    idxBad = np.logical_or(idxBad, np.logical_or(x_too_big, y_too_big))
    x_too_small = posdata[NotNaN,1] < -max_error
    y_too_small = posdata[NotNaN,2] < -max_error
    idxBad = np.logical_or(idxBad, np.logical_or(x_too_small, y_too_small))
    # Combine good position data lines
    posdata = np.delete(posdata, NotNaN[idxBad], axis=0)

    return posdata


def remove_tracking_data_jumps(posdata, maxjump):
    """
    Removes data with too large jumps based on euclidean distance

    posdata - numpy array with columns:
              timestamps
              LED 1 xpos
              LED 1 ypos
              LED 2 xpos
              LED 2 ypos
              , where NaN for missing LED 2 data
    maxjump - int or float specifying maximum allowed shift in euclidean distance
    """
    keepPos = []
    lastPos = posdata[0,1:3]
    for npos in range(posdata.shape[0]):
        currpos = posdata[npos,1:3]
        if euclidean(lastPos, currpos) < maxjump:
            keepPos.append(npos)
            lastPos = currpos
    keepPos = np.array(keepPos)
    print(str(posdata.shape[0] - keepPos.size) + ' of ' + 
          str(posdata.shape[0]) + ' removed in postprocessing')
    posdata = posdata[keepPos,:]

    return posdata


def iteratively_combine_multicamera_data_for_recording(
    CameraSettings, arena_size, posdatas, OE_GC_times, verbose=False):
    """Return ProcessedPos

    Combines raw tracking data from multiple cameras into a single ProcessedPos array.
    If a single camera data is provided, this will be converted into same format.
    """
    cameraIDs = sorted(CameraSettings['CameraSpecific'].keys())
    # Load position data for all cameras
    for i, posdata in enumerate(posdatas):
        if isinstance(posdata, dict):
            # This is conditional because for early recordings tracking data was immediately stored in Open Ephys time.
            # Remove datapoints where all position data is None
            idx_None = np.all(np.isnan(posdata['OnlineTrackerData']), 1)
            posdata['OnlineTrackerData'] = np.delete(posdata['OnlineTrackerData'], 
                                                     np.where(idx_None)[0], axis=0)
            posdata['OnlineTrackerData_timestamps'] = np.delete(posdata['OnlineTrackerData_timestamps'], 
                                                                np.where(idx_None)[0], axis=0)
            # Compute position data timestamps in OpenEphys time
            RPi_frame_times = posdata['OnlineTrackerData_timestamps']
            RPi_GC_times = posdata['GlobalClock_timestamps']
            RPi_frame_in_OE_times = estimate_OpenEphys_timestamps_for_tracking_data(
                OE_GC_times, RPi_GC_times, RPi_frame_times, verbose=verbose
            )
            # Combine timestamps and position data into a single array
            posdata = np.concatenate((RPi_frame_in_OE_times.astype(np.float64)[:, None], posdata['OnlineTrackerData']), axis=1)
            posdatas[i] = posdata
    if len(posdatas) > 1:
        # If data from multiple cameras available, combine it
        PosDataFramesPerSecond = 30.0
        # Find first and last timepoint for position data
        first_timepoints = []
        last_timepoints = []
        for posdata in posdatas:
            first_timepoints.append(posdata[0, 0])
            last_timepoints.append(posdata[-1, 0])
        # Combine position data step-wise from first to last timepoint at PosDataFramesPerSecond
        # At each timepoint the closest matchin datapoints will be taken from different cameras
        timepoints = np.arange(np.array(first_timepoints).min(), np.array(last_timepoints).max(), 1.0 / PosDataFramesPerSecond)
        combPosData = [None]
        listNaNs = []
        if verbose:
            print('Combining iteratively camera data for each position timepoint')
        for npoint in (tqdm(range(len(timepoints))) if verbose else range(len(timepoints))):
            # Find closest matchin timepoint from all RPis
            idx_tp = np.zeros(4, dtype=np.int32)
            for nRPi in range(len(posdatas)):
                idx_tp[nRPi] = np.argmin(np.abs(timepoints[npoint] - posdatas[nRPi][:,0]))
            # Convert posdatas for use in combineCamerasData function
            cameraPos = []
            for nRPi in range(len(posdatas)):
                cameraPos.append(posdatas[nRPi][idx_tp[nRPi], 1:5])
            tmp_comb_data = combineCamerasData(cameraPos, combPosData[-1], cameraIDs, CameraSettings, arena_size)
            combPosData.append(tmp_comb_data)
            if tmp_comb_data is None:
                listNaNs.append(npoint)
        # Remove the extra element from combPosData
        del combPosData[0]
        # Remove all None elements
        for nanElement in listNaNs[::-1]:
            del combPosData[nanElement]
        timepoints = np.delete(timepoints, listNaNs)
        # Combine timepoints and position data
        ProcessedPos = np.concatenate((np.expand_dims(np.array(timepoints), axis=1), np.array(combPosData)), axis=1)
        # Print info about None elements
        if len(listNaNs) > 0:
            print('Total of ' + str(len(listNaNs) * (1.0 / PosDataFramesPerSecond)) + ' seconds of position data was lost')
            if listNaNs[0] == 0:
                print('This was in the beginning and ' + str(np.sum(np.diff(np.array(listNaNs)) > 1)) + ' other epochs.')
            else:
                print('This was not in the beginning, but in ' + str(np.sum(np.diff(np.array(listNaNs)) > 1) + 1) + ' other epochs.')
    else:
        # In case of single camera being used, just use data from that camera
        ProcessedPos = posdatas[0]
    ProcessedPos = remove_tracking_data_outside_boundaries(ProcessedPos, arena_size, max_error=10)
    ProcessedPos = ProcessedPos.astype(np.float64)

    return ProcessedPos



import os
import cv2
import argparse
import numpy as np

os.sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import NWBio
from TrackingDataProcessing import estimate_OpenEphys_timestamps_for_tracking_data

def get_video_resolution(filename):
    '''
    Return frame resolution of a video file.
    '''
    reader = cv2.VideoCapture(filename)
    ret = False
    while not ret:
        ret, frame = reader.read()
    resolution = (frame.shape[1], frame.shape[0])
    reader.release()

    return resolution

def get_video_timestamps(filename, cameraID):
    posdata = NWBio.load_raw_tracking_data(filename, cameraID)
    # Compute position data timestamps in OpenEphys time
    RPi_frame_times = posdata['VideoData_timestamps']
    RPi_GC_times = posdata['GlobalClock_timestamps']
    OE_GC_times = NWBio.load_GlobalClock_timestamps(filename)
    RPi_frame_in_OE_times = estimate_OpenEphys_timestamps_for_tracking_data(OE_GC_times, RPi_GC_times, RPi_frame_times)

    return RPi_frame_in_OE_times

def get_pos_data(filename, cameraID):
    posdata = NWBio.load_raw_tracking_data(filename, cameraID)
    idx_None = np.all(np.isnan(posdata['OnlineTrackerData']), 1)
    posdata['OnlineTrackerData'] = np.delete(posdata['OnlineTrackerData'], 
                                             np.where(idx_None)[0], axis=0)
    posdata['OnlineTrackerData_timestamps'] = np.delete(posdata['OnlineTrackerData_timestamps'], 
                                                        np.where(idx_None)[0], axis=0)
    # Compute position data timestamps in OpenEphys time
    RPi_frame_times = posdata['OnlineTrackerData_timestamps']
    RPi_GC_times = posdata['GlobalClock_timestamps']
    if not ('OE_GC_times' in locals()):
        OE_GC_times = NWBio.load_GlobalClock_timestamps(filename)
    RPi_frame_in_OE_times = estimate_OpenEphys_timestamps_for_tracking_data(OE_GC_times, RPi_GC_times, RPi_frame_times)
    # Combine timestamps and position data into a single array
    posdata = np.concatenate((RPi_frame_in_OE_times.astype(np.float64)[:, None], posdata['OnlineTrackerData']), axis=1)

    return posdata

class video_read_write(object):
    '''
    Makes video file frames sequentially available via read() method and
    can write frames sequentially into a new video MP4 file via write() method.
    '''
    def __init__(self, filename, framerate, new_file_suffix='_annotated', frame_limit=None):
        '''
        filename - full path to video file to read
        framerate - framerate of output video file in Hz
        new_file_suffix (optional) - prefix appended to new video file
        '''
        if not isinstance(new_file_suffix, str) or len(new_file_suffix) == 0:
            raise ValueError('new_file_suffix must be a string of length > 0.')
        self.frame_limit = frame_limit
        self.__frame_pos = -1
        resolution = get_video_resolution(filename)
        self.reader = cv2.VideoCapture(filename)
        out_filename = os.path.splitext(os.path.basename(filename))[0] + new_file_suffix + '.mp4'
        out_file_path = os.path.join(os.path.dirname(filename), out_filename)
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        self.writer = cv2.VideoWriter(out_file_path, fourcc, framerate, resolution)

    @property
    def frame_pos(self):
        '''
        The position of last frame returned from read() method. Count starts from 0.
        '''
        return self.__frame_pos

    def read(self):
        '''
        Returns the next frame. If no more frames available, returns None.
        '''
        ret, frame = self.reader.read()
        if ret:
            self.__frame_pos += 1
            if self.frame_limit is None or self.__frame_pos < self.frame_limit:
                return frame

    def write(self, frame):
        self.writer.write(frame)

    def close(self):
        self.reader.release()
        self.writer.release()


class video_annotator(object):
    '''
    Rewrites a video file, annotating each frame iteratively with annotator classes. 
    '''
    def __init__(self, annotator_classes, filename, framerate, new_file_suffix='_annotated', frame_limit=None):
        '''
        annotator_classes - iterable of classes to use for annotating each frame.
                            See annotate() method for futher description annotator_classes.
                            Note! close method is called on each class if available.
        filename - full path to video file to read
        framerate - framerate of output video file in Hz
        new_file_suffix (optional) - prefix appended to new video file
        '''
        self.video_files = video_read_write(filename, framerate, new_file_suffix, frame_limit=frame_limit)
        self.annotator_classes = annotator_classes
        self.annotate()
        self.close()

    def annotate(self):
        '''
        Iterates over all frames in the video file and each frame is modified
        by iterating over annotator_classes and calling the process() method of
        each class on the frame. Inputs to process() method are (frame ,frame counter).
        Frame counter starts from 0 on first frame.
        '''
        frame = self.video_files.read()
        while not (frame is None):
            for annotator in self.annotator_classes:
                frame = annotator.process(frame, self.video_files.frame_pos)
            self.video_files.write(frame)
            frame = self.video_files.read()

    def close(self):
        self.video_files.close()
        for annotator in self.annotator_classes:
            if hasattr(annotator, 'close'):
                annotator.close()


class annotator_position(object):
    '''
    Draws circles to locations where position was detected in tracking data
    '''
    def __init__(self, filename, cameraID, lag_signal_threshold=0.5):
        '''
        filename - str - full path to NWB file.
        cameraID - str - ID of camera data to use for annotating images.
        lag_signal_threshold - float - when closest matching position data timepoint
                               is off by more than lag_signal_threshold seconds,
                               the marker is made smaller and less bright.
        '''
        self.lag_signal_threshold = lag_signal_threshold
        resolution_option = NWBio.load_settings(filename, '/CameraSettings/General/resolution_option/')
        calibrationTmatrix = NWBio.load_settings(filename, '/CameraSettings/CameraSpecific/' + \
                                                           cameraID + '/CalibrationData/' + resolution_option + \
                                                           '/calibrationTmatrix')
        self.frame_timestamps = get_video_timestamps(filename, cameraID)
        self.calibrationTmatrix = np.linalg.inv(calibrationTmatrix)
        posdata = get_pos_data(filename, cameraID)
        self.pos_timestamps = posdata[:,0]
        self.pos_xys = (posdata[:,1:3], posdata[:,3:5])
        self.colors = ((255, 0, 0),  (0, 255, 0))
        self.widths = (4, 2)
        self.radius = 10

    def get_color_and_width(self, n_xy, distance):
        color = self.colors[n_xy]
        width = self.widths[n_xy]
        if distance > self.lag_signal_threshold:
            color = tuple(map(int, np.uint8(np.array(color) / 2.0)))
            width = int(width / 2.0)

        return color, width

    def process(self, frame, frame_pos):
        timestamp = self.frame_timestamps[frame_pos]
        idx = (np.abs(self.pos_timestamps - timestamp)).argmin()
        distance = np.abs(timestamp - self.pos_timestamps[idx])
        for n_xy in range(len(self.pos_xys)):
            xy = self.pos_xys[n_xy][idx, :]
            if not np.isnan(xy[0]):
                tmp = np.reshape(np.array([xy[0], xy[1]],dtype=np.float32),(1,1,2))
                tmp = cv2.perspectiveTransform(tmp, self.calibrationTmatrix)
                xy = tuple(list(np.int64(np.round(tmp.squeeze()))))
                color, width = self.get_color_and_width(n_xy, distance)
                frame = cv2.circle(frame, xy, self.radius, color, width)

        return frame


def annotate_with_position(folder_path, cameraID, framerate, frame_limit=None):
    filename = NWBio.get_filename(folder_path)
    pos_annotator = annotator_position(filename, cameraID)
    video_file_name = NWBio.load_raw_tracking_data(filename, cameraID, 
                                                   specific_path='VideoFile')
    video_file_path = os.path.join(folder_path, video_file_name)
    video_annotator((pos_annotator,), video_file_path, framerate, frame_limit=frame_limit)


def main(args):
    folder_path = args.folder_path[0]
    cameraID = args.cameraID[0]
    if args.framerate:
        framerate = args.framerate[0]
    else:
        framerate = 30
    if args.frame_limit:
        frame_limit = args.frame_limit[0]
    else:
        frame_limit = None
    annotate_with_position(folder_path, cameraID, framerate, frame_limit=frame_limit)

if __name__ == '__main__':
    # Input argument handling and help info
    parser = argparse.ArgumentParser(description='Convert all jpg images in folder to mp4 video.')
    parser.add_argument('folder_path', type=str, nargs=1, 
                        help='Path to recording folder.')
    parser.add_argument('cameraID', type=str, nargs=1, 
                        help='Camera ID.')
    parser.add_argument('--framerate', type=int, nargs = 1, 
                        help='Frame rate of the output video (Hz).')
    parser.add_argument('--frame_limit', type=int, nargs = 1, 
                        help='Maxium number of frames to process.')
    args = parser.parse_args()
    main(args)




from PyQt5 import QtWidgets
from PyQt5.QtCore import QTimer
from functools import partial
import numpy as np
from copy import copy
from openEPhys_DACQ import NWBio
from openEPhys_DACQ.HelperFunctions import openSingleFileDialog, show_message, QThread_with_completion_callback
import pyqtgraph
from openEPhys_DACQ.RPiInterface import CameraControl
from io import BytesIO
from socket import socket
import struct
from PIL import Image
from threading import Thread, Lock
import cv2


class CameraStreamCapture(object):
    """
    Listens to MPJEG stream from camera and prepares images for display.
    """
    def __init__(self, stream_port, update_frame, cameraID):
        """
        stream_port - int - port at 0.0.0.0 at which to listen to the MJPEG stream
        update_frame - is called with after receiving each frame with cameraID and prepared frame
        cameraID - key for images_for_display and image_for_display_Locks correct value
        """
        self.update_frame = update_frame
        self.cameraID = cameraID
        # Set internal variables
        self.calibration_overlay_on = False
        self.replacement_frame = None
        self.prepare_frame_Lock = Lock()
        # Start thread for streaming
        self.stream_open = True
        self.T_stream = Thread(target=self._stream, args=(stream_port,))
        self.T_stream.start()

    def _overlay_calibration_on_frame(self, frame):
        """
        Adds overlay onto the frame.
        """
        pattern = self.calibration_pattern
        ndots_x = self.calibration_ndots_xy[0]
        ndots_y = self.calibration_ndots_xy[1]
        frame = cv2.drawChessboardCorners(frame, (ndots_x, ndots_y), pattern, True)

        return frame

    def _prepare_frame(self, image_stream):
        """
        Prepares frame from MJPEG stream as a numpy array or uses the replacement frame
        and adds overlay of calibration if requested.
        Calls update_frame provided at initialization with self.cameraID and prepared frame.
        """
        with self.prepare_frame_Lock:
            if self.replacement_frame is None:
                # Rewind the stream, open it as an image with PIL and convert to numpy
                image_stream.seek(0)
                image = Image.open(image_stream)
                frame = np.array(image)
            else:
                frame = copy(self.replacement_frame)
            # Overlay calibration pattern if requested
            if self.calibration_overlay_on:
                frame = self._overlay_calibration_on_frame(frame)
        # Update images_for_display
        self.update_frame(self.cameraID, frame)

    def _stream(self, port):
        """
        Listens to MJPEG stream at 0.0.0.0 specified port.
        Calls _prepare_frame() method with fileld BytesIO class for each received frame.
        Finishes if self.stream_open is False or stream is closed.
        """
        # Open connection
        self.server_socket = socket()
        self.server_socket.bind(('0.0.0.0', int(port)))
        self.server_socket.listen(0)
        self.connection = self.server_socket.accept()[0].makefile('rb')
        try:
            while self.stream_open:
                # Read the length of the image as a 32-bit unsigned int. If the
                # length is zero, quit the loop
                image_len = struct.unpack('<L', self.connection.read(struct.calcsize('<L')))[0]
                if not image_len:
                    break
                # Construct a stream to hold the image data and read the image
                # data from the connection
                image_stream = BytesIO()
                image_stream.write(self.connection.read(image_len))
                Thread(target=self._prepare_frame, args=(image_stream,)).start()
        except Exception as e:
            if not isinstance(e, struct.error):
                raise e

    def calibration_overlay(self, on_off, pattern=None, ndots_xy=None):
        """
        Sets calibration overlay.

        on_off - bool
        pattern - output from CameraRPiController.Calibrator.get_calibration_data
        ndots_xy - [int - dots in X axis, int - dots in Y axis]
        """
        with self.prepare_frame_Lock:
            self.calibration_pattern = pattern
            self.calibration_ndots_xy = ndots_xy
            self.calibration_overlay_on = on_off

    def use_replacement_frame(self, on_off, frame=None):
        """
        Sets replacement frame.
        on_off - bool
        frame - replacement frame as numpy array - required if on_off=True
        """
        with self.prepare_frame_Lock:
            if on_off:
                self.replacement_frame = copy(frame)
            else:
                self.replacement_frame = None

    def close(self):
        self.stream_open = False
        self.T_stream.join()
        self.connection.close()
        self.server_socket.close()


class CameraSettingsApp(object):
    """
    Handles camera settings, calibration and provides live stream of camera feed.

    Note! CameraSettingsApp requires a specific workflow for calibration data to match the settings. 
          All settings must be specified before any other methods are called.
    """
    def __init__(self, apply_settings_method, camera_settings=None):
        """
        :param apply_settings_method: this called by :py:func:`CameraSettingsApp.apply` with camera_settings as arg
        :param dict camera_settings: camera_settings to initialise the applicaiton with. Otherwise defaults are used.
        """
        # Default variables
        self.display_stream_ports_start = 8000
        # Create internal variables
        self.CameraControllers = {}
        self.CameraStreamCaptures = {}
        self.CalibrationData = {}
        # Initialize
        self.apply_settings_method = apply_settings_method
        self.general_settings = self.default_settings('general_settings')
        self.camera_specific_settings = self.default_settings('camera_specific_settings')
        if not (camera_settings is None):
            self.import_settings(camera_settings['General'], camera_settings['CameraSpecific'])

    @staticmethod
    def _default_general_settings():
        """
        Specifies default settigs as well as data types.
        """
        return {'tracking_mode':          {'value': 'dual_led', 
                                           'description': ('Tracking mode', ('1 LED', 'single_led'), 
                                                                            ('2 LED', 'dual_led'), 
                                                                            ('Motion', 'motion'))}, 
                'resolution_option':      {'value': 'low', 
                                           'description': ('Video resolution', ('Low', 'low'), ('High', 'high'))}, 
                'LED_separation':         {'value': np.float64(5.0), 
                                           'description': 'LED separation (cm)'}, 
                'LED_angle':              {'value': np.float64(0.0), 
                                           'description': 'LED angle'}, 
                'camera_transfer_radius': {'value': np.float64(40.0), 
                                           'description': 'Camera transfer radius (cm)'}, 
                'smoothing_box':          {'value': np.int64(5), 
                                           'description': 'OnlineTracker smoothing (pix)'}, 
                'motion_threshold':       {'value': np.int64(10), 
                                           'description': '[Motion] Detection threshold (0-255)'}, 
                'motion_size':            {'value': np.int64(500), 
                                           'description': '[Motion] Pixel count threshold'}, 
                'framerate':              {'value': np.int64(30), 
                                           'description': 'Frame rate (Hz)'}, 
                'centralIP':              {'value': '192.168.0.10', 
                                           'description': 'IP address of Recording PC'}, 
                'global_clock_ip':        {'value': '192.168.0.8', 
                                           'description': 'IP address of Global Clock RPi'}, 
                'password':               {'value': 'raspberry', 
                                           'description': 'Camera RPi password'}, 
                'username':               {'value': 'pi', 
                                           'description': 'Camera RPi username'}, 
                'OnlineTracker_port':     {'value': '5600', 
                                           'description': 'OnlineTracker port'}, 
                'ZMQcomms_port':          {'value': '5601', 
                                           'description': 'ZMQcomms port'}
                }

    @staticmethod
    def _default_camera_specific_settings():
        """
        Specifies default settigs as well as data types.
        """
        return {'1': {'address':  {'value': '192.168.0.20', 
                                   'description': 'IP'}, 
                      'location_xy': {'value': np.array([350.0, 250.0], dtype=np.float64), 
                                   'description': 'Location X,Y (cm)'}, 
                      'offset_xy':   {'value': np.array([6.0, 6.0], dtype=np.float64), 
                                   'description': 'Offset X,Y (cm)'}, 
                      'spacing':  {'value': np.float64(20.38), 
                                   'description': 'Spacing (cm)'}, 
                      'ndots_xy':  {'value': np.array([4, 11], dtype=np.int64), 
                                   'description': 'Dots X,Y'}
                      }
                }

    def general_settings_keyorder(self, settings=None):
        if settings is None:
            settings = self.general_settings
        if not (settings is None):
            return sorted(settings.keys())

    def single_camera_specific_settings_keyorder(self, settings=None):
        if settings is None:
            cameraID = list(self.camera_specific_settings.keys())[0]
            settings = self.camera_specific_settings[cameraID]
        if not (settings is None):
            return sorted(settings.keys())

    def default_settings(self, settings_name):
        if settings_name == 'general_settings':
            return CameraSettingsApp._default_general_settings()
        elif settings_name == 'camera_specific_settings':
            return CameraSettingsApp._default_camera_specific_settings()

    def default_single_camera_specific_settings(self):
        """
        Returns settings and keyorder of the first cameraID
        """
        camera_specific_settings = self.default_settings('camera_specific_settings')
        settings = camera_specific_settings[list(camera_specific_settings.keys())[0]]
        keyorder = self.single_camera_specific_settings_keyorder()

        return settings

    def update_settings(self, general_settings, camera_specific_settings):
        self.general_settings = general_settings
        self.camera_specific_settings = camera_specific_settings

    def import_settings(self, general_settings, camera_specific_settings):
        """
        Creates new settings dictionaries by copying values into default settings.
        """
        new_general_settings = self.default_settings('general_settings')
        for key in new_general_settings.keys():
            if key in general_settings.keys():
                new_general_settings[key]['value'] = general_settings[key]
        new_camera_specific_settings = {}
        for cameraID in list(camera_specific_settings.keys()):
            if 'CalibrationData' in camera_specific_settings[cameraID].keys():
                self.CalibrationData[cameraID] = camera_specific_settings[cameraID].pop('CalibrationData')
            new_camera_specific_settings[cameraID] = self.default_single_camera_specific_settings()
            for key in new_camera_specific_settings[cameraID].keys():
                if key in camera_specific_settings[cameraID].keys():
                    new_camera_specific_settings[cameraID][key]['value'] = camera_specific_settings[cameraID][key]
        self.general_settings = new_general_settings
        self.camera_specific_settings = new_camera_specific_settings

    def export_settings(self):
        """
        Creates a simplified dictionary of settings that is compatible with import_settings() method.
        """
        general = {}
        for key in self.general_settings.keys():
            general[key] = self.general_settings[key]['value']
        camera_specific = {}
        for cameraID in self.camera_specific_settings.keys():
            camera_specific[cameraID] = {}
            for key in self.camera_specific_settings[cameraID].keys():
                camera_specific[cameraID][key] = self.camera_specific_settings[cameraID][key]['value']
            if cameraID in self.CalibrationData.keys():
                camera_specific[cameraID]['CalibrationData'] = self.CalibrationData[cameraID]

        return {'General': general, 'CameraSpecific': camera_specific}

    def load_from_file(self, filename):
        CameraSettings = NWBio.load_settings(filename, '/CameraSettings/')
        self.import_settings(CameraSettings['General'], CameraSettings['CameraSpecific'])

    def save_to_file(self, filename):
        NWBio.save_settings(filename, self.export_settings(), path='/CameraSettings/')

    def apply(self):
        """
        Calls the apply_settings_method provided at initialization with export_settings() method output.
        """
        self.apply_settings_method(self.export_settings())

    def _initialize_camera(self, cameraID):
        """
        Initializes control of a specific camera.
        """
        address = self.camera_specific_settings[cameraID]['address']['value']
        port = self.general_settings['ZMQcomms_port']['value']
        username = self.general_settings['username']['value']
        password = self.general_settings['password']['value']
        resolution_option = 'low'
        framerate = 10
        self.CameraControllers[cameraID] = CameraControl(address, port, username, 
                                                         password, resolution_option, 
                                                         framerate=framerate)

    def initialize_cameras(self):
        """
        Initializes control of all cameras according to current camera_specific_settings.
        """
        self.CameraControllers = {}
        t__initialize_camera = []
        for cameraID in self.camera_specific_settings.keys():
            t = Thread(target=self._initialize_camera, args=(cameraID,))
            t.start()
            t__initialize_camera.append(t)
        for t in t__initialize_camera:
            t.join()

    def _close_camera(self, cameraID):
        """
        Closes specific camera controller.
        """
        controller = self.CameraControllers.pop(cameraID)
        controller.close()

    def close_cameras(self):
        """
        Closes all camera controllers.
        """
        t_close_camera = []
        for key in list(self.CameraControllers.keys()):
            t = Thread(target=self._close_camera, args=(key,))
            t.start()
            t_close_camera.append(t)
        for t in t_close_camera:
            t.join()
        del self.CameraControllers

    def ensure_cameras_are_closed(self):
        if hasattr(self, 'CameraControllers'):
            self.close_cameras()

    def camera_display_start(self, update_frame):
        """
        Starts CameraStreamCapture for all cameras.

        update_frame - is called with after receiving each frame with cameraID and prepared frame
        """
        # Produce separate ports for streaming from each camera
        display_stream_ports = {}
        for n_cam, cameraID in enumerate(self.camera_specific_settings.keys()):
            display_stream_ports[cameraID] = self.display_stream_ports_start + n_cam
        # Start CameraStreamCapture class for each camera
        for cameraID in self.camera_specific_settings.keys():
            self.CameraStreamCaptures[cameraID] = CameraStreamCapture(display_stream_ports[cameraID], 
                                                                      update_frame, cameraID)
        # Start streaming from all cameras
        for cameraID in self.camera_specific_settings.keys():
            self.CameraControllers[cameraID].start_streaming(self.general_settings['centralIP']['value'], 
                                                             display_stream_ports[cameraID])

    def camera_display_stop(self):
        """
        Stops CameraStreamCapture of all cameras.
        """
        for key in list(self.CameraStreamCaptures.keys()):
            controller = self.CameraStreamCaptures.pop(key)
            controller.close()

    def initialize_cameras_and_streaming(self, update_frame):
        """
        First initializes cameras and then starts CameraStreamCapture for all cameras.

        update_frame - is called with after receiving each frame with cameraID and prepared frame
        """
        self.initialize_cameras()
        self.camera_display_start(update_frame)

    def stop_cameras_and_streaming(self):
        """
        Stops camera controllers and then CameraStreamCapture of all cameras.
        """
        self.close_cameras()
        self.camera_display_stop()

    def calibrate(self, cameraID):
        """
        Erases any previous calibration data.

        Attempts to calibrate the camera with current settings.
        If successful, the data is stored in self.CalibrationData.

        Returns True if calibration successful, False otherwise.

        cameraID - specifies which camera to calibrate.
        """
        if cameraID in list(self.CalibrationData.keys()):
            _ = self.CalibrationData.pop(cameraID)
        if cameraID in self.CameraControllers.keys():
            calibration_parameters = {'ndots_xy': self.camera_specific_settings[cameraID]['ndots_xy']['value'], 
                                      'spacing': self.camera_specific_settings[cameraID]['spacing']['value'], 
                                      'offset_xy': self.camera_specific_settings[cameraID]['offset_xy']['value']}
            calibration = self.CameraControllers[cameraID].calibrate(calibration_parameters)
            if not (calibration is None):
                self.CalibrationData[cameraID] = calibration
                return True
            else:
                return False

    def overlay_calibration(self, cameraID, on_off):
        """
        Sets calibration overlay if CalibrationData available for cameraID.
        """
        if (cameraID in self.CalibrationData.keys()) and (cameraID in self.CameraStreamCaptures.keys()):
            if on_off:
                pattern = self.CalibrationData[cameraID]['low']['pattern']
                ndots_xy = self.camera_specific_settings[cameraID]['ndots_xy']['value']
                self.CameraStreamCaptures[cameraID].calibration_overlay(True, pattern, ndots_xy)
            else:
                self.CameraStreamCaptures[cameraID].calibration_overlay(False)

    def show_calibration_frame(self, cameraID, on_off):
        """
        Sets live view replacement with calibration image, 
        if CalibrationData available for cameraID.
        """
        if (cameraID in self.CalibrationData.keys()) and (cameraID in self.CameraStreamCaptures.keys()):
            if on_off:
                frame = self.CalibrationData[cameraID]['low']['frame']
                self.CameraStreamCaptures[cameraID].use_replacement_frame(True, frame)
            else:
                self.CameraStreamCaptures[cameraID].use_replacement_frame(False)

    def close(self):
        self.ensure_cameras_are_closed()
        if hasattr(self, 'CameraStreamCaptures'):
            for cameraID in self.CameraStreamCaptures.keys():
                self.CameraStreamCaptures[cameraID].close()
        if hasattr(self, 'GUI'):
            self.GUI.close()


class radio_button_menu(QtWidgets.QWidget):

    def __init__(self, name_value_pairs):
        super(radio_button_menu, self).__init__()
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0,0,0,0)
        self.buttons = {}
        for name, value in name_value_pairs:
            self.buttons[value] = QtWidgets.QRadioButton(name)
            layout.addWidget(self.buttons[value])
        self.setLayout(layout)

    def set_checked(self, value):
        self.buttons[value].setChecked(True)

    def get_checked(self):
        for value in self.buttons.keys():
            if self.buttons[value].isChecked():
                return value


def setDoubleBoxStretch(box, first, second):
    box.setStretch(0, first)
    box.setStretch(1, second)
    return box


class SettingsMenu(QtWidgets.QWidget):
    """
    Widget with multiple editable settings.
    """
    def __init__(self, settings, keyorder, primary_orientation, secondary_orientation):
        """
        settings - dict - from CameraSettings.default_settings()
        keyorder - list - settings.keys() in preferred order
        primary_orientation - str - 'horizontal' or 'vertical' or 'h' or 'v'
                              orientation of the main layout
        secondary_orientation - str - 'horizontal' or 'vertical' or 'h' or 'v'
                                orientation of each individual settings layout
        """
        super(SettingsMenu, self).__init__()
        self.settings = settings
        # Create top layout
        if primary_orientation == 'horizontal' or primary_orientation == 'h':
            layout = QtWidgets.QHBoxLayout()
        elif primary_orientation == 'vertical' or primary_orientation == 'v':
            layout = QtWidgets.QVBoxLayout()
        else:
            raise ValueError('primary_orientation argument must be horizontal or vertical')
        layout.setContentsMargins(0,0,0,0)
        self.setLayout(layout)
        # Create settings option
        self.handles = {}
        for key in keyorder:
            if secondary_orientation == 'horizontal' or secondary_orientation == 'h':
                box = QtWidgets.QHBoxLayout()
            elif secondary_orientation == 'vertical' or secondary_orientation == 'v':
                box = QtWidgets.QVBoxLayout()
            else:
                raise ValueError('secondary_orientation argument must be horizontal or vertical')
            box.setContentsMargins(0,0,0,0)
            # Create settings option type based on settings[key]['description value']
            if isinstance(self.settings[key]['description'], str):
                box.addWidget(QtWidgets.QLabel(self.settings[key]['description']))
                self.handles[key] = QtWidgets.QLineEdit()
            elif isinstance(self.settings[key]['description'], tuple):
                box.addWidget(QtWidgets.QLabel(self.settings[key]['description'][0]))
                self.handles[key] = radio_button_menu(self.settings[key]['description'][1:])
            box.addWidget(self.handles[key])
            # Add individual setting to SettingsMenu layout in a frame
            frame = QtWidgets.QFrame()
            frame.setLayout(setDoubleBoxStretch(box, 2, 1))
            frame.setFrameStyle(0)
            layout.addWidget(frame)
        # Populate SettingsMenu values with data from settings
        self.put(self.settings)

    def put(self, settings):
        """
        Populates SettingsMenu in handles with data from settings.

        settings - dict - with structure corresponding to SettingsMenu() input.
        """
        for key in settings.keys():
            value = copy(settings[key]['value'])
            if isinstance(self.handles[key], QtWidgets.QLineEdit):
                if isinstance(value, np.ndarray):
                    value = ','.join(map(str, value))
                elif isinstance(value, np.int64) or isinstance(value, np.float64):
                    value = str(value)
                elif not isinstance(value, str):
                    raise ValueError('Expected numpy.ndarray, numpy.int64, numpy.float64 or str,\n' + \
                                     'but got ' + str(type(value)))
                self.handles[key].setText(value)
            elif isinstance(self.handles[key], radio_button_menu):
                self.handles[key].set_checked(value)
            else:
                raise ValueError('Expected ' + str(QtWidgets.QLineEdit) + ' or ' + str(radio_button_menu) +',\n' + \
                                 'but got ' + str(type(self.handles[key])))

    def get(self):
        """
        Returns settings values with values extracted from SettingsMenu in handles.
        """
        for key in self.settings.keys():
            if isinstance(self.handles[key], QtWidgets.QLineEdit):
                value = str(self.handles[key].text())
                if isinstance(self.settings[key]['value'], np.ndarray):
                    self.settings[key]['value'] = np.array(value.split(','), 
                                                           dtype=self.settings[key]['value'].dtype)
                elif isinstance(self.settings[key]['value'], np.int64):
                    self.settings[key]['value'] = np.int64(value)
                elif isinstance(self.settings[key]['value'], np.float64):
                    self.settings[key]['value'] = np.float64(value)
                elif isinstance(self.settings[key]['value'], str):
                    self.settings[key]['value'] = value
                else:
                    raise ValueError('Expected numpy.ndarray, numpy.int64, numpy.float64 or str,\n' + \
                                     'but got ' + str(type(self.settings[key]['value'])))
            elif isinstance(self.handles[key], radio_button_menu):
                self.settings[key]['value'] = self.handles[key].get_checked()
            else:
                raise ValueError('Expected ' + str(QtWidgets.QLineEdit) + ' or ' + str(radio_button_menu) +',\n' + \
                                 'but got ' + str(type(self.handles[key])))

        return copy(self.settings)


class CameraDisplayWidget(QtWidgets.QWidget):
    """
    Widget that displays images from cameras in a grid.

    Use update_frame() method to update to make new frames available to be displayed.
    """
    def __init__(self, cameraIDs, fps=10):
        """
        cameraIDs - list - cameraID values to use for accessing individual displays
        fps - int - image update frequency
        """
        super(CameraDisplayWidget, self).__init__()
        self.cameraIDs = cameraIDs
        # Compute viewbox locations and set layout
        view_pos = CameraDisplayWidget.assign_viewbox_locations(len(self.cameraIDs))
        layout = QtWidgets.QGridLayout()
        self.setLayout(layout)
        # Prepare variables
        self.images = {}
        self.image_Locks = {}
        self.viewboxes = {}
        self.imageItems = {}
        self.timers = {}
        # Populate variables for each camera
        for npos, cameraID in enumerate(self.cameraIDs):
            # Create image and lock
            self.images[cameraID] = np.uint8(np.random.random((300, 300, 3)) * 255)
            self.image_Locks[cameraID] = Lock()
            # Create viewbox
            graphics_view = pyqtgraph.GraphicsView()
            graphics_layout = pyqtgraph.GraphicsLayout()
            graphics_layout.addLabel('Camera ' + str(cameraID), col=0, row=0)
            graphics_layout.setContentsMargins(0, 0, 0, 0)
            self.viewboxes[cameraID] = graphics_layout.addViewBox(col=0, row=1)
            self.viewboxes[cameraID].setAspectLocked(True)
            self.viewboxes[cameraID].invertY()
            self.imageItems[cameraID] = pyqtgraph.ImageItem()
            self.viewboxes[cameraID].addItem(self.imageItems[cameraID])
            graphics_view.setCentralItem(graphics_layout)
            layout.addWidget(graphics_view, view_pos['row'][npos], view_pos['col'][npos])
            # Start updating image in viewbox
            self.timers[cameraID] = QTimer()
            self.timers[cameraID].timeout.connect(partial(self._display_frame, cameraID))
            self.timers[cameraID].start(int(1000 / fps))

    @staticmethod
    def assign_viewbox_locations(n_cameras):
        """
        Returns the number of columns and rows necessary for number of cameras
        with columns having the priority.
        """
        n_cols = 1
        n_rows = 1
        increase_next = 'cols'
        while n_rows * n_cols < n_cameras:
            if increase_next == 'cols':
                n_cols += 1
                increase_next = 'rows'
            elif increase_next == 'rows':
                n_rows += 1
                increase_next = 'cols'
        # Assign grid positions for each camera
        view_pos = {'col': [], 'row': []}
        for npos in range(n_cameras):
            view_pos['col'].append(int(np.mod(npos, n_cols)))
            view_pos['row'].append(int(np.floor(npos / float(n_cols))))

        return view_pos

    def update_frame(self, cameraID, frame):
        """
        Updates the current latest frame for camera cameraID.

        frame - numpy array of shape (height, width, 3)
        """
        with self.image_Locks[cameraID]:
            self.images[cameraID] = frame

    def _display_frame(self, cameraID):
        """
        Uses self.images and self.image_Locks to acquire current image.
        Then displays it in the correct viewbox.
        """
        with self.image_Locks[cameraID]:
            frame = self.images[cameraID]
        frame = np.swapaxes(frame, 0, 1)
        self.imageItems[cameraID].setImage(frame)

    def close(self, *args, **kwargs):
        for cameraID in self.timers.keys():
            self.timers[cameraID].stop()
        super(CameraDisplayWidget, self).close(*args, **kwargs)



class CameraSettingsGUI(QtWidgets.QDialog):
    """
    GUI for CameraSettingsApp.
    """
    def __init__(self, CameraSettingsApp, parent=None):
        """
        :param CameraSettingsApp CameraSettingsApp:
        """
        super(CameraSettingsGUI, self).__init__(parent=parent)
        self.CameraSettingsApp = CameraSettingsApp
        # Initialize GUI window
        self.resize(1100, 750)
        self.setWindowTitle('Camera Settings')
        # Create top menu items
        self.loadButton = QtWidgets.QPushButton('Load')
        self.loadButton.clicked.connect(self.loadButton_click)
        self.saveButton = QtWidgets.QPushButton('Save')
        self.saveButton.clicked.connect(self.saveButton_click)
        self.apply_button = QtWidgets.QPushButton('Apply')
        self.apply_button.clicked.connect(self.apply_button_callback)
        self.cancel_button = QtWidgets.QPushButton('Cancel')
        self.cancel_button.clicked.connect(self.cancel_button_callback)
        self.addCameraButton = QtWidgets.QPushButton('Add Camera')
        self.addCameraButton.clicked.connect(self.addCameraButton_click)
        self.initializeCamerasButton = QtWidgets.QPushButton('Init. Cameras')
        self.initializeCamerasButton.setCheckable(True)
        self.initializeCamerasButton.clicked.connect(self.initializeCamerasButton_click)
        self.testTrackingButton = QtWidgets.QPushButton('Test Tracking')
        self.testTrackingButton.clicked.connect(self.testTrackingButton_click)
        top_menu_widget = QtWidgets.QWidget()
        top_menu_layout = QtWidgets.QVBoxLayout(top_menu_widget)
        top_menu_layout.addWidget(self.loadButton)
        top_menu_layout.addWidget(self.saveButton)
        top_menu_layout.addWidget(self.apply_button)
        top_menu_layout.addWidget(self.cancel_button)
        top_menu_layout.addWidget(self.addCameraButton)
        top_menu_layout.addWidget(self.initializeCamerasButton)
        top_menu_layout.addWidget(self.testTrackingButton)
        top_menu_frame = QtWidgets.QFrame()
        top_menu_frame.setLayout(top_menu_layout)
        top_menu_frame.setFrameStyle(3)
        # Create Camera Specific Settings menu items
        self.camera_specific_items = {}
        self.camera_specific_items_key_counter = 0
        camera_specific_settings_master_widget = QtWidgets.QWidget()
        self.camera_specific_settings_layout = QtWidgets.QVBoxLayout(camera_specific_settings_master_widget)
        # Create General Settings menu layout in a frame
        self.general_SettingsMenu_layout = QtWidgets.QVBoxLayout()
        general_SettingsMenu_frame = QtWidgets.QFrame()
        general_SettingsMenu_frame.setLayout(self.general_SettingsMenu_layout)
        general_SettingsMenu_frame.setFrameStyle(0)
        general_settings_scroll_area = QtWidgets.QScrollArea()
        general_settings_scroll_area.setWidget(general_SettingsMenu_frame)
        general_settings_scroll_area.setWidgetResizable(True)
        # Create General Settings menu scroll area
        camera_specific_settings_scroll_area = QtWidgets.QScrollArea()
        camera_specific_settings_scroll_area.setWidget(camera_specific_settings_master_widget)
        camera_specific_settings_scroll_area.setWidgetResizable(True)
        # Create Camera View Box
        self.camera_display_layout = QtWidgets.QVBoxLayout()
        camera_display_frame = QtWidgets.QFrame()
        camera_display_frame.setLayout(self.camera_display_layout)
        camera_display_frame.setFrameStyle(3)
        camera_display_frame.setMinimumSize(700,500)
        # Put all boxes into main window
        top_hbox = QtWidgets.QHBoxLayout()
        top_hbox.addWidget(top_menu_frame)
        top_hbox.addWidget(camera_specific_settings_scroll_area)
        bottom_hbox = QtWidgets.QHBoxLayout()
        bottom_hbox.addWidget(general_settings_scroll_area)
        bottom_hbox.addWidget(camera_display_frame)
        main_vbox = QtWidgets.QVBoxLayout()
        main_vbox.addItem(top_hbox)
        main_vbox.addItem(setDoubleBoxStretch(bottom_hbox, 1, 2))
        self.setLayout(setDoubleBoxStretch(main_vbox, 1, 2))
        # Load settings into SettingsMenus
        self.write_settings_into_GUI(self.CameraSettingsApp.general_settings, 
                                     self.CameraSettingsApp.general_settings_keyorder(),
                                     self.CameraSettingsApp.camera_specific_settings, 
                                     self.CameraSettingsApp.single_camera_specific_settings_keyorder())
        self.show()

    def _add_camera_specific_settings(self, settings, keyorder):
        """
        Creates camera_specific_settings and adds it to the camera_specific_settings_layout.

        settings - as required by SettingsMenu class
        keyorder - as required by SettingsMenu class
        """
        key = copy(self.camera_specific_items_key_counter)
        self.camera_specific_items_key_counter += 1
        self.camera_specific_items[key] = {}
        single_camera_settings_layout = QtWidgets.QHBoxLayout()
        single_camera_settings_layout.setContentsMargins(2,2,2,2)
        # Add all settings
        self.camera_specific_items[key]['SettingsMenu'] = SettingsMenu(settings, keyorder, 'h', 'v')
        single_camera_settings_layout.addWidget(self.camera_specific_items[key]['SettingsMenu'])
        # Add Buttons
        buttons_layout = QtWidgets.QGridLayout()
        buttons_layout.setContentsMargins(0,0,0,0)
        single_camera_settings_layout.addItem(buttons_layout)
        # Add button to remove this camera
        self.camera_specific_items[key]['removeButton'] = QtWidgets.QPushButton('Remove')
        self.camera_specific_items[key]['removeButton'].clicked.connect(lambda: self._remove_single_camera_settings(key))
        buttons_layout.addWidget(self.camera_specific_items[key]['removeButton'], 0, 0)
        # Add button to calibrate this camera
        self.camera_specific_items[key]['calibrationButton'] = QtWidgets.QPushButton('Calibrate')
        self.camera_specific_items[key]['calibrationButton'].clicked.connect(lambda: self.calibrationButton_click(key))
        self.camera_specific_items[key]['calibrationButton'].setEnabled(False)
        buttons_layout.addWidget(self.camera_specific_items[key]['calibrationButton'], 0, 1)
        # Add button to show calibration image for this camera
        self.camera_specific_items[key]['showCalibrationButton'] = QtWidgets.QPushButton('Show')
        self.camera_specific_items[key]['showCalibrationButton'].clicked.connect(lambda: self.showCalibrationButton_click(key))
        self.camera_specific_items[key]['showCalibrationButton'].setEnabled(False)
        self.camera_specific_items[key]['showCalibrationButton'].setCheckable(True)
        buttons_layout.addWidget(self.camera_specific_items[key]['showCalibrationButton'], 1, 0)
        # Add button to overlay calibration of this camera
        self.camera_specific_items[key]['calibrationOverlayButton'] = QtWidgets.QPushButton('Overlay')
        self.camera_specific_items[key]['calibrationOverlayButton'].clicked.connect(lambda: self.calibrationOverlayButton_click(key))
        self.camera_specific_items[key]['calibrationOverlayButton'].setEnabled(False)
        self.camera_specific_items[key]['calibrationOverlayButton'].setCheckable(True)
        buttons_layout.addWidget(self.camera_specific_items[key]['calibrationOverlayButton'], 1, 1)
        # Create frame for single camera_specific_settings
        single_camera_settings_frame = QtWidgets.QFrame()
        single_camera_settings_frame.setLayout(single_camera_settings_layout)
        single_camera_settings_frame.setFrameStyle(2)
        single_camera_settings_frame.setMaximumHeight(60)
        self.camera_specific_settings_layout.addWidget(single_camera_settings_frame)
        self.camera_specific_items[key]['frame_remover'] = single_camera_settings_frame.deleteLater

    def _remove_single_camera_settings(self, key):
        """
        Removes single camera specific settings from GUI
        """
        self.camera_specific_items[key]['SettingsMenu'].close()
        self.camera_specific_items[key]['removeButton'].close()
        self.camera_specific_items[key]['calibrationButton'].close()
        self.camera_specific_items[key]['showCalibrationButton'].close()
        self.camera_specific_items[key]['calibrationOverlayButton'].close()
        self.camera_specific_items[key]['frame_remover']()
        del self.camera_specific_items[key]

    def _clear_camera_specific_settings_in_GUI(self):
        """
        Removes all camera specific settings from GUI
        """
        for key in list(self.camera_specific_items.keys()):
            self._remove_single_camera_settings(key)

    def _get_cameraID_from_camera_specific_items(self, key):
        """
        Returns edited cameraID from camera specific SettingsMenu
        """
        settings = self.camera_specific_items[key]['SettingsMenu'].get()
        return settings['ID']['value']

    def _read_settings_from_GUI(self):
        """
        Returns general and camera specific settings from SettingsMenus in format
        compatible with CameraSettingsApp.update_settings.
        """
        general_settings = self.general_SettingsMenu.get()
        camera_specific_settings = {}
        cameraIDs = []
        for key in self.camera_specific_items.keys():
            settings = self.camera_specific_items[key]['SettingsMenu'].get()
            cameraID = settings.pop('ID')['value']
            cameraIDs.append(cameraID)
            camera_specific_settings[cameraID] = settings
        if len(cameraIDs) != len(set(cameraIDs)):
            show_message('Duplicates not allowed in camera IDs.', 
                         'Currently set IDs: ' + str(sorted(cameraIDs)))
            raise ValueError('Duplicates not allowed in camera IDs.\n' + \
                             'Currently set IDs: ' + str(sorted(cameraIDs)))

        return general_settings, camera_specific_settings

    def write_settings_into_GUI(self, general_settings, general_settings_keyorder, 
                                camera_specific_settings, single_camera_specific_settings_keyorder):
        """
        Updates general and camera specific settings in GUI

        general_settings - CameraSettingsApp.general_settings value
        general_settings_keyorder - display keyorder for general_settings
        camera_specific_settings - CameraSettingsApp.camera_specific_settings value
        single_camera_specific_settings_keyorder - display keyorder for camera_specific_settings
        """
        if hasattr(self, 'general_SettingsMenu'):
            self.general_SettingsMenu.deleteLater()
            del self.general_SettingsMenu
        self.general_SettingsMenu = SettingsMenu(general_settings, general_settings_keyorder, 'v', 'h')
        self.general_SettingsMenu_layout.addWidget(self.general_SettingsMenu)
        self._clear_camera_specific_settings_in_GUI()
        keyorder = self.CameraSettingsApp.single_camera_specific_settings_keyorder()
        keyorder = ['ID'] + keyorder
        for cameraID in sorted(camera_specific_settings.keys()):
            settings = copy(camera_specific_settings[cameraID])
            settings['ID'] = {'value': cameraID, 'description': 'ID'}
            self._add_camera_specific_settings(settings, keyorder)

    def _update_settings_on_CameraSettingsApp(self):
        """
        Calls CameraSettingsApp.update_settings with output from _read_settings_from_GUI() method.
        """
        general_settings, camera_specific_settings = self._read_settings_from_GUI()
        self.CameraSettingsApp.update_settings(general_settings, camera_specific_settings)

    def loadButton_click(self):
        """
        Attempts to load settings into CameraSettingsApp from selected file.
        """
        filename = openSingleFileDialog('load', suffix='nwb', caption='Select file')
        if not (filename is None):
            self.CameraSettingsApp.load_from_file(filename)
            self.write_settings_into_GUI(self.CameraSettingsApp.general_settings,
                                         self.CameraSettingsApp.general_settings_keyorder(),
                                         self.CameraSettingsApp.camera_specific_settings,
                                         self.CameraSettingsApp.single_camera_specific_settings_keyorder())

    def saveButton_click(self):
        """
        Attempts to save settings from CameraSettingsApp to selected file.
        """
        self._update_settings_on_CameraSettingsApp()
        filename = openSingleFileDialog('save', suffix='nwb', caption='Save file name and location')
        if not (filename is None):
            self.CameraSettingsApp.save_to_file(filename)

    def apply_button_callback(self):
        """
        Calls CameraSettingsApp.apply method.
        """
        self._update_settings_on_CameraSettingsApp()
        self.CameraSettingsApp.apply()
        self.CameraSettingsApp.close()
        self.close()

    def cancel_button_callback(self):
        """
        Calls CameraSettingsApp.close, which in turn calls the close() method in CameraSettingsGUI.
        """
        self.CameraSettingsApp.close()
        self.close()

    def addCameraButton_click(self):
        """
        Calls _add_camera_specific_settings() method with 
        default_single_camera_specific_settings from CameraSettingsApp.
        """
        default_cameraID = '0'
        settings = self.CameraSettingsApp.default_single_camera_specific_settings()
        keyorder = self.CameraSettingsApp.single_camera_specific_settings_keyorder()
        settings['ID'] = {'value': default_cameraID, 'description': 'ID'}
        keyorder = ['ID'] + keyorder
        self._add_camera_specific_settings(settings, keyorder)

    def initializeCamerasButton_click(self):
        """
        Enables and disables correct functions on GUI.
        If toggled on:
            Initializes CameraDisplayWidget
            calls _update_settings_on_CameraSettingsApp() method
            calls CameraSettingsApp.initialize_cameras_and_streaming
        If toggled off:
            calls CameraSettingsApp.stop_cameras_and_streaming
            calls CameraDisplayWidget.close
        """
        # Disable or enable GUI interactions
        if not hasattr(self, 'initializeCamerasButtonDefaultStyleSheet'):
            self.initializeCamerasButtonDefaultStyleSheet = self.initializeCamerasButton.styleSheet()
        initializeCamerasButton_state = self.initializeCamerasButton.isChecked()
        self.initializeCamerasButton.setEnabled(False)
        self.addCameraButton.setEnabled(not initializeCamerasButton_state)
        self.general_SettingsMenu.setEnabled(not initializeCamerasButton_state)
        for key in self.camera_specific_items.keys():
            self.camera_specific_items[key]['SettingsMenu'].setEnabled(not initializeCamerasButton_state)
            self.camera_specific_items[key]['removeButton'].setEnabled(not initializeCamerasButton_state)
            self.camera_specific_items[key]['calibrationButton'].setEnabled(initializeCamerasButton_state)
            self.camera_specific_items[key]['showCalibrationButton'].setEnabled(initializeCamerasButton_state)
            self.camera_specific_items[key]['calibrationOverlayButton'].setEnabled(initializeCamerasButton_state)
        if initializeCamerasButton_state:
            self.initializeCamerasButton.setStyleSheet('background-color: red')
            # Initialize CameraDisplayWidget class based on camera_specific_settings
            general_settings, camera_specific_settings = self._read_settings_from_GUI()
            cameraIDs = camera_specific_settings.keys()
            self.CameraDisplayWidget = CameraDisplayWidget(sorted(cameraIDs))
            self.camera_display_layout.addWidget(self.CameraDisplayWidget)
            # Call camera_display_start() method on CameraSettingsApp after updating CameraSettings
            self._update_settings_on_CameraSettingsApp()
            self.main_worker = QThread_with_completion_callback(self._initializeCameras_finished_callback, 
                                                                self.CameraSettingsApp.initialize_cameras_and_streaming, 
                                                                function_args=(self.CameraDisplayWidget.update_frame,))
        else:
            # Close CameraDisplayWidget class
            self.CameraSettingsApp.stop_cameras_and_streaming()
            self.CameraDisplayWidget.close()
            self.initializeCamerasButton.setStyleSheet(self.initializeCamerasButtonDefaultStyleSheet)
            self.initializeCamerasButton.setEnabled(True)

    def _initializeCameras_finished_callback(self):
        """
        Is called once CameraSettingsApp.initialize_cameras_and_streaming call is finished.
        Enables initializeCamerasButton and sets it green.
        """
        self.initializeCamerasButton.setEnabled(True)
        self.initializeCamerasButton.setStyleSheet('background-color: green')

    def calibrationButton_click(self, key):
        """
        Calls CameraSettingsApp.calibrate for the correct cameraID.
        """
        self.camera_specific_items[key]['calibrationButton'].setEnabled(False)
        if not hasattr(self, 'calibrationButtonDefaultStyleSheet'):
            self.calibrationButtonDefaultStyleSheet = self.camera_specific_items[key]['calibrationButton'].styleSheet()
        self.camera_specific_items[key]['calibrationButton'].setStyleSheet('background-color: red')
        cameraID = self._get_cameraID_from_camera_specific_items(key)
        self.main_worker = QThread_with_completion_callback(self.calibration_finished_callback, 
                                                            self.CameraSettingsApp.calibrate, 
                                                            return_output=True, 
                                                            callback_args=(key,), 
                                                            function_args=(cameraID,))


    def calibration_finished_callback(self, outcome, key):
        """
        Is called once CameraSettingsApp.calibrate call is finished.
        Enables calibrationButton and sets it green or default, depending on outcome.
        """
        self.camera_specific_items[key]['calibrationButton'].setEnabled(True)
        if outcome == True:
            self.camera_specific_items[key]['calibrationButton'].setStyleSheet('background-color: green')
        else:
            self.camera_specific_items[key]['calibrationButton'].setStyleSheet(self.calibrationButtonDefaultStyleSheet)

    def showCalibrationButton_click(self, key):
        """
        Calls CameraSettingsApp.show_calibration_frame.
        """
        isChecked = self.camera_specific_items[key]['showCalibrationButton'].isChecked()
        cameraID = self._get_cameraID_from_camera_specific_items(key)
        Thread(target=self.CameraSettingsApp.show_calibration_frame, 
                   args=(cameraID, isChecked)).start()

    def calibrationOverlayButton_click(self, key):
        """
        Calls CameraSettingsApp.overlay_calibration.
        """
        isChecked = self.camera_specific_items[key]['calibrationOverlayButton'].isChecked()
        cameraID = self._get_cameraID_from_camera_specific_items(key)
        Thread(target=self.CameraSettingsApp.overlay_calibration, 
                   args=(cameraID, isChecked)).start()

    def testTrackingButton_click(self):
        raise NotImplementedError




import os
import sys
from copy import deepcopy
import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets
from uuid import uuid4
from datetime import datetime
from time import sleep
import threading
from importlib import import_module
import multiprocessing

from openEPhys_DACQ.package_configuration import package_config
from openEPhys_DACQ import HelperFunctions as HFunc
from openEPhys_DACQ import RPiInterface
from openEPhys_DACQ import NWBio
from openEPhys_DACQ.CameraSettings import CameraSettingsApp, CameraSettingsGUI
from openEPhys_DACQ.TaskSettings import TaskSettingsApp, TaskSettingsGUI
from openEPhys_DACQ.ZMQcomms import SubscribeToOpenEphys, PublishToOpenEphys
from openEPhys_DACQ.CumulativePosPlot import PosPlot


def find_latest_time_folder(path):
    """Looks up subdirectory names and returns folder with latest date time string as folder name
    """
    dir_items = os.listdir(path)
    dir_times = []
    item_nrs = []
    for n_item in range(len(dir_items)):
        try:  # Use only items for which the name can be converted to correct date time format
            item_time = datetime.strptime(dir_items[n_item], '%Y-%m-%d_%H-%M-%S')
            item_nrs.append(n_item)
            dir_times.append(item_time)
        except:
            pass
    # Find the latest time based on folder names
    latest_time = max(dir_times)
    latest_folder = dir_items[item_nrs[dir_times.index(latest_time)]]

    return latest_folder


def get_recording_file_path(recording_folder_root):
    recording_file_name = 'experiment_1.nwb'
    if os.path.isdir(recording_folder_root):
        # Get the name of the folder with latest date time as name
        latest_folder = find_latest_time_folder(recording_folder_root)
        folder_timestamp = datetime.strptime(latest_folder, '%Y-%m-%d_%H-%M-%S')
        # If the folder time is less than 5 seconds old
        if (datetime.now() - folder_timestamp).total_seconds() < 5:
            recording_file = os.path.join(recording_folder_root, latest_folder, recording_file_name)
        else:  # Display error in text box
            recording_file = False
    else:  # Display error in text box
        recording_file = False

    return recording_file


def check_if_nwb_recording(fpath):
    if os.path.isfile(fpath):
        # Check for file size at intervals to see if it is changing in size. Stop checking if constant.
        previous_size = os.stat(fpath).st_size
        sleep(0.1)
        current_size = os.stat(fpath).st_size
        file_recording = previous_size != current_size
    else:
        file_recording = False

    return file_recording


def store_camera_data_to_recording_file(camera_settings, fpath):
    # Initialize file_manager for each camera
    file_managers = {}
    for cameraID in camera_settings['CameraSpecific'].keys():
        print(['DEBUG', HFunc.time_string(), 'Initializing Camera_RPi_file_manager for ID: ' + cameraID])
        address = camera_settings['CameraSpecific'][cameraID]['address']
        username = camera_settings['General']['username']
        file_managers[cameraID] = RPiInterface.Camera_RPi_file_manager(address, username)
        print(
            ['DEBUG', HFunc.time_string(), 'Initializing Camera_RPi_file_manager for ID: ' + cameraID + ' complete.'])
    # Retrieve data concurrently from all cameras
    thread_list = []
    for cameraID in camera_settings['CameraSpecific'].keys():
        t = threading.Thread(target=file_managers[cameraID].retrieve_timestamps_and_OnlineTrackerData)
        t.start()
        thread_list.append(t)
    for t in thread_list:
        t.join()
    # Combine data from cameras and store to recording file
    camera_data = {}
    for cameraID in camera_settings['CameraSpecific'].keys():
        print(['DEBUG', HFunc.time_string(), 'Retrieving camera_data for ID: ' + cameraID])
        camera_data[cameraID] = file_managers[cameraID].get_timestamps_and_OnlineTrackerData(cameraID)
        print(['DEBUG', HFunc.time_string(), 'Retrieving camera_data for ID: ' + cameraID + ' complete.'])
    print(['DEBUG', HFunc.time_string(), 'Saving camera_data to recording file'])
    NWBio.save_tracking_data(fpath, camera_data)
    print(['DEBUG', HFunc.time_string(), 'Saving camera_data to recording file complete.'])
    # Copy video data directly to recording folder
    thread_list = []
    for cameraID in camera_settings['CameraSpecific'].keys():
        t = threading.Thread(target=file_managers[cameraID].copy_over_video_data,
                             args=(os.path.dirname(fpath), cameraID))
        t.start()
        thread_list.append(t)
    for t in thread_list:
        t.join()


def list_general_settings_history(path):
    """
    Assumes all files on the path are NWB files with General settings stored

    Returns:
        dictionary - just as General settings, but each key contains
                     a list of values for that key in files on the path.
        list       - full paths to all settings files

        All lists are sorted starting from the most recent timestamp in filename.
        None is entered if key is missing in settings file.
    """
    dir_items = os.listdir(path)
    filetimes = []
    filenames = []
    for item in dir_items:
        if item.endswith('.settings.nwb'):
            filename = os.path.join(path, item)
            filenames.append(filename)
            filetimes.append(datetime.strptime(item[:19], '%Y-%m-%d_%H-%M-%S'))
    # Sort filenames based on timestamps
    filenames = [x for _, x in sorted(zip(filetimes, filenames))][::-1]
    # Load all general settings to memory and build a list of keys
    settings_keys = []
    settings_list = []
    for filename in filenames:
        settings = NWBio.load_settings(filename, path='/General/')
        for key in settings.keys():
            if not (key in settings_keys):
                settings_keys.append(key)
        settings_list.append(settings)
    # Create lists for all keys
    general_settings_history = {}
    for key in settings_keys:
        general_settings_history[key] = []
    for settings in settings_list:
        for key in settings_keys:
            if key in settings.keys():
                general_settings_history[key].append(settings[key])
            else:
                general_settings_history[key].append(None)

    return general_settings_history, filenames


class RecordingManagerException(Exception):
    pass


class OpenEphysMessenger(object):

    def __init__(self):

        self._closed = False

        self._pipe, self.open_ephys_message_pipe = multiprocessing.Pipe()

        print('Connecting to Open Ephys GUI via ZMQ...')
        self._subscriber = SubscribeToOpenEphys(verbose=False)
        self._subscriber.connect()
        self._publisher = PublishToOpenEphys()
        print('Connecting to Open Ephys GUI via ZMQ Successful')

        self._subscriber.add_callback(lambda msg: self._pipe.send(msg))
        self._publisher_thread = threading.Thread(target=self.publisher_thread_method)
        self._publisher_thread.start()

    def send_message_to_open_ephys(self, msg):
        self._publisher.sendMessage(msg)

    @property
    def closed(self):
        return self._closed

    def publisher_thread_method(self):
        while not self.closed:
            if self._pipe.poll(0.1):
                self._publisher.sendMessage(self._pipe.recv())

    def close(self):
        self._subscriber.disconnect()
        self._publisher.close()
        self._closed = True
        self._publisher_thread.join()


def format_channel_map(channel_map):
    if not isinstance(channel_map, dict):
        raise ValueError('channel_map must be dict')
    for channel_group in channel_map:
        if not isinstance(channel_group, str):
            raise ValueError('Channel map fields must be strings')
        channel_map[channel_group] = {
            'string': str(channel_map[channel_group]['string']),
            'list': np.array(channel_map[channel_group]['list'])
        }

    return channel_map


class RecordingManager(object):
    """
    Central recording controller. Provides interface to specify settings and start/stop the recording.
    """

    general_settings = {
        'experimenter': '',
        'root_folder': package_config()['root_folder'],
        'animal': '',
        'experiment_id': '',
        'arena_size': np.array([87.5, 125]),
        'badChan': '',
        'rec_file_path': '',
        'Tracking': np.array(False),
        'channel_map': {},
        'TaskActive': np.array(False)
    }

    general_settings_formatting = {
        'experimenter': str,
        'root_folder': str,
        'animal': str,
        'experiment_id': str,
        'arena_size': np.array,
        'badChan': str,
        'rec_file_path': str,
        'Tracking': np.array,
        'channel_map': format_channel_map,
        'TaskActive': np.array,
    }

    camera_settings = None

    task_settings = None

    settings_ready = {
        'general_settings': True,
        'camera_settings': False,
        'task_settings': False
    }

    recording_initialized = False
    recording_active = False
    recording_closing = False

    RecordingManagerSettingsFolder = os.path.join(package_config()['root_folder'],
                                                  package_config()['recording_manager_settings_subfolder'])

    global_clock_controller = None
    tracking_controller = None
    online_tracking_processor = None
    open_ephys_messenger = None
    current_task = None
    position_plot = None

    def update_general_settings(self, key, value):

        if key in ('experimenter', 'root_folder', 'animal', 'experiment_id', 'badChan', 'rec_file_path'):
            assert isinstance(value, str)
        elif key in ('Tracking', 'TaskActive'):
            assert isinstance(value, bool) or (isinstance(value, np.ndarray) and value.dtype == np.bool)
        elif key in ('arena_size',):
            assert isinstance(value, (list, np.ndarray)) and len(value) == 2
        elif key in ('channel_map',):
            assert isinstance(value, dict)
        else:
            raise ValueError('key {} not found.'.format(key))

        self.general_settings[key] = self.general_settings_formatting[key](value)

        if key == 'arena_size':
            self.settings_ready['task_settings'] = False

    def update_camera_settings(self, camera_settings):
        self.camera_settings = camera_settings
        self.settings_ready['camera_settings'] = True

    def update_task_settings(self, task_settings):
        self.task_settings = task_settings
        self.settings_ready['task_settings'] = True

    def camera_settings_app(self):
        return CameraSettingsApp(self.update_camera_settings, self.camera_settings)

    def task_settings_app(self):
        return TaskSettingsApp(self.general_settings['arena_size'], self.update_task_settings,
                               self.task_settings)

    def save_settings(self, fpath):
        """Saves settings into NWB file.
        """
        NWBio.save_settings(fpath, {'General': self.general_settings,
                                    'Time': datetime.now().strftime('%Y-%m-%d_%H-%M-%S'),
                                    'CameraSettings': self.camera_settings,
                                    'TaskSettings': self.task_settings
                                    })

    def load_settings(self, fpath):
        """Loads settings from NWB file.
        """
        settings = NWBio.load_settings(fpath)
        if 'General' in settings:
            for key, value in settings['General'].items():
                self.update_general_settings(key, value)
        if 'CameraSettings' in settings:
            self.update_camera_settings(settings['CameraSettings'])
        if 'TaskSettings' in settings:
            self.update_task_settings(settings['TaskSettings'])

    @staticmethod
    def find_latest_matching_settings_filepath(path, settings=None):

        general_settings_history, filenames = list_general_settings_history(path)

        if len(filenames) == 0:

            return None

        if settings is None or len(settings.keys()) == 0:

            # If no specific settings were requested, return the path to latest settings file
            filepath = filenames[0]

        else:

            # Loop through all settings files
            filepaths = []
            for nfile in range(len(filenames)):
                # Check if all specified settings match to the one in the file
                settings_correct = []
                for key in settings.keys():
                    settings_correct.append(general_settings_history[key][nfile] == settings[key])
                if all(settings_correct):
                    filepaths.append(filenames[nfile])
            if len(filepaths) > 0:  # Use the most recent settings file path
                filepath = filepaths[0]
            else:  # If no settings matched, return None
                filepath = None

        return filepath

    def list_general_setting_options(self, setting_key):
        general_settings_history, _ = list_general_settings_history(self.RecordingManagerSettingsFolder)
        return sorted(list(set(general_settings_history[setting_key])))

    def load_last_settings(self):
        # Hard-code which settings to use for searching the latest settings
        settings = {}
        if len(self.general_settings['animal']) > 0:
            settings['animal'] = self.general_settings['animal']
        if len(self.general_settings['experiment_id']) > 0:
            settings['experiment_id'] = self.general_settings['experiment_id']
        # Find the latest settings file
        fpath = self.find_latest_matching_settings_filepath(self.RecordingManagerSettingsFolder, settings)
        if fpath is None:
            # If no matching file is found, show error message
            criteria = ''
            for key in sorted(settings.keys()):
                criteria += '{}: {}\n'.format(key, settings[key])

            raise ValueError('No settings found for criteria:\n' + criteria)
        else:
            self.load_settings(fpath)

    @staticmethod
    def test_ping_of_ip_addresses(ip_addresses):
        """Returns a list of devices that fail to ping back.

        :param tuple ip_addresses: tuple of list of IP address strings
        :return: tuple of bool values for each element of ip_addresses. True if device can be pinged successfully.
        """

        def test_ping_and_update_locked_list(ip_address, locked_list, list_pos, list_lock):
            ret = HFunc.test_pinging_address(ip_address)
            with list_lock:
                locked_list[list_pos] = ret

        # Check each device in a seperate thread to speed things up
        output_lock = threading.Lock()
        address_ping_outcome = [None] * len(ip_addresses)
        threads_address_pings = []
        for n_address, address in enumerate(ip_addresses):
            t = threading.Thread(target=test_ping_and_update_locked_list,
                                 args=[address, address_ping_outcome, n_address, output_lock])
            t.start()
            threads_address_pings.append(t)
        for t in threads_address_pings:
            t.join()

        return tuple(address_ping_outcome)

    def test_devices(self):
        """Returns status of devices required for recording.

        :return: list of dicts {'name': str, 'available': bool}
        """

        ip_addresses = []
        device_names = []

        # Add tracking cameras to list
        if self.general_settings['Tracking']:
            if self.settings_ready['camera_settings']:
                assert not (self.camera_settings is None)  # camera_settings should not be None if settings_ready

                for cameraID in sorted(self.camera_settings['CameraSpecific'].keys(), key=int):
                    ip_addresses.append(self.camera_settings['CameraSpecific'][cameraID]['address'])
                    device_names.append('Camera {}'.format(cameraID))

            else:

                raise RecordingManagerException('Tracking is turned on, but Camera Settings are not available.')

        # Add feeders to list
        if self.general_settings['TaskActive']:
            if self.settings_ready['task_settings']:
                assert not (self.task_settings is None)  # task_settings should not be None if settings_ready

                for feeder_type in self.task_settings['FEEDERs'].keys():
                    for feeder_id in sorted(self.task_settings['FEEDERs'][feeder_type].keys(), key=int):

                        if self.task_settings['FEEDERs'][feeder_type][feeder_id]['Active']:

                            ip_addresses.append(self.task_settings['FEEDERs'][feeder_type][feeder_id]['IP'])
                            device_names.append('Feeder {} {}'.format(feeder_type, feeder_id))

            else:

                raise RecordingManagerException('Task is turned on, but Task Settings are not available.')

        device_names = tuple(device_names)

        # Test pinging of these devices
        device_availability = self.test_ping_of_ip_addresses(tuple(ip_addresses))

        return tuple([{'name': name,
                       'available': availability}
                      for name, availability in zip(device_names, device_availability)])

    def check_if_ready_to_initialize(self):

        if self.general_settings['Tracking']:
            if not self.settings_ready['camera_settings']:
                raise RecordingManagerException('Tracking is turned on, but Camera Settings are not available.')

        if self.general_settings['TaskActive']:
            if not self.general_settings['Tracking']:
                raise RecordingManagerException('Tracking must be turned on')
            if not self.settings_ready['task_settings']:
                raise RecordingManagerException('Task is turned on, but Task Settings are not available.')

    def create_task_arguments(self):
        return (
            deepcopy(self.task_settings),
            self.open_ephys_messenger.open_ephys_message_pipe,
            self.online_tracking_processor.combPosHistory,
            self.online_tracking_processor.combPos_update_interval,
            self.online_tracking_processor.position_histogram_dict
        )

    def init_devices(self):

        if self.recording_active or self.recording_closing:
            raise RecordingManagerException('Recording already ongoing.')

        if self.recording_initialized:
            raise RecordingManagerException('Recording already initialized.')

        self.check_if_ready_to_initialize()

        # Initialize tracking
        if self.general_settings['Tracking']:

            print('Connecting to GlobalClock RPi...')
            args = (self.camera_settings['General']['global_clock_ip'],
                    self.camera_settings['General']['ZMQcomms_port'],
                    self.camera_settings['General']['username'],
                    self.camera_settings['General']['password'])
            self.global_clock_controller = RPiInterface.GlobalClockControl(*args)
            print('Connecting to GlobalClock RPi Successful')

            # Connect to tracking RPis
            print('Connecting to tracking RPis...')
            self.tracking_controller = RPiInterface.TrackingControl(self.camera_settings)
            print('Connecting to tracking RPis Successful')

            # Initialize onlineTrackingData class
            print('Initializing Online Tracking Data Processor...')
            self.online_tracking_processor = RPiInterface.onlineTrackingData(
                self.camera_settings, self.general_settings['arena_size'],
                HistogramParameters={'margins': 10,  # histogram data margins in centimeters
                                     'binSize': 2,  # histogram binSize in centimeters
                                     'speedLimit': 10}  # centimeters of distance in last second to be included
            )
            print('Initializing Tracking Data Processor Successful')

        # Initialize talking to OpenEphysGUI
        self.open_ephys_messenger = OpenEphysMessenger()

        # Initialize Task
        if self.general_settings['TaskActive']:

            print('Initializing Task...')
            task_module = import_module('Tasks.' + self.task_settings['name'] + '.Task')
            self.current_task = task_module.Core(*self.create_task_arguments())
            print('Initializing Task Successful')

        self.recording_initialized = True
        print('Recording Initialization Successful')

    def cancel_init_devices(self):

        if not self.recording_initialized:
            raise RecordingManagerException('Devices are not initialized.')

        # Initialize tracking
        if self.general_settings['Tracking']:

            print('Closing GlobalClock Controller...')
            self.global_clock_controller.close()
            print('Closing GlobalClock Controller Successful')

            print('Closing tracking RPis...')
            self.tracking_controller.close()
            print('Closing tracking RPis Successful')

            print('Closing Online Tracking Data...')
            self.online_tracking_processor.close()
            print('Closing Online Tracking Data Successful')

        # Close channel to OpenEphysGUI
        self.open_ephys_messenger.close()

        # Initialize Task
        if self.general_settings['TaskActive']:

            print('Closing Task...')
            self.current_task.stop()
            print('Closing Task Successful')

        self.recording_initialized = False
        print('Recording Initialization Cancelled successfully')

    def start_rec(self):

        if self.recording_active or self.recording_closing:
            raise RecordingManagerException('Recording already ongoing.')

        if not self.recording_initialized:
            raise RecordingManagerException('Devices must be initialized first')
        else:
            self.recording_initialized = False

        # Start Open Ephys GUI recording
        print('Starting Open Ephys GUI Recording...')
        recording_folder_root = os.path.join(self.general_settings['root_folder'], self.general_settings['animal'])
        command = 'StartRecord RecDir=' + recording_folder_root + ' CreateNewDir=1'
        self.open_ephys_messenger.send_message_to_open_ephys(command)

        # Make sure OpenEphys is recording
        recording_file = get_recording_file_path(recording_folder_root)
        while not recording_file:
            sleep(0.1)
            recording_file = get_recording_file_path(recording_folder_root)
        while not check_if_nwb_recording(recording_file):
            sleep(0.1)
        self.general_settings['rec_file_path'] = recording_file
        print('Starting Open Ephys GUI Recording Successful')

        # Start the tracking scripts on all RPis
        if self.general_settings['Tracking']:
            print('Starting tracking RPis...')
            self.tracking_controller.start()
            print('Starting tracking RPis Successful')
            # Tracking controller start command should not complete before RPis have really executed the command
            print('Starting GlobalClock RPi...')
            self.global_clock_controller.start()
            print('Starting GlobalClock RPi Successful')

        # Start cumulative plot
        if self.general_settings['Tracking']:
            print('Starting Position Plot...')
            self.position_plot = PosPlot(self.online_tracking_processor, self.camera_settings['General']['LED_angle'])
            print('Starting Position Plot Successful')

        # Start task
        if self.general_settings['TaskActive']:
            print('Starting Task...')
            self.current_task.run()
            print('Starting Task Successful')

        self.recording_active = True

    def compile_recording_data(self):

        # Store recording settings to recording file
        self.save_settings(self.general_settings['rec_file_path'])
        print('Settings saved to Recording File')

        # Store settings for RecordingManager history reference
        RecordingManagerSettingsPath = os.path.join(self.general_settings['root_folder'],
                                                    self.RecordingManagerSettingsFolder)
        if not os.path.isdir(RecordingManagerSettingsPath):
            os.mkdir(RecordingManagerSettingsPath)
        RecordingFolderName = os.path.basename(os.path.dirname(self.general_settings['rec_file_path']))
        RecordingManagerSettingsFilePath = os.path.join(RecordingManagerSettingsPath,
                                                        RecordingFolderName + '.settings.nwb')
        self.save_settings(RecordingManagerSettingsFilePath)
        print('Settings saved to Recording Manager Settings Folder')

        if self.general_settings['Tracking']:
            # Store tracking  data in recording file
            print('Copying over tracking data to Recording File...')
            store_camera_data_to_recording_file(self.camera_settings, self.general_settings['rec_file_path'])
            print('Copying over tracking data to Recording File Successful')

    def stop_rec(self):

        if self.recording_closing:
            raise RecordingManagerException('Recording closing already in progress.')
        elif not self.recording_active or self.recording_initialized:
            raise RecordingManagerException('Recording is not currently ongoing.')
        else:
            self.recording_active = False
            self.recording_closing = True

        # Stop Task
        if self.general_settings['TaskActive']:
            print('Stopping Task...')
            self.current_task.stop()
            print('Stopping Task Successful')

        # Stop updating tracking system
        if self.general_settings['Tracking']:
            print('Closing Online Tracking Data...')
            self.online_tracking_processor.close()
            print('Closing Online Tracking Data Successful')
            print('Closing GlobalClock Controller...')
            self.global_clock_controller.stop()
            self.global_clock_controller.close()
            print('Closing GlobalClock Controller Successful')
            # Stop cameras
            print('Stopping tracking RPis...')
            self.tracking_controller.close()
            print('Stopping tracking RPis Successful')

        # Stop Open Ephys Recording
        while check_if_nwb_recording(self.general_settings['rec_file_path']):
            print('Stopping Open Ephys GUI Recording...')
            self.open_ephys_messenger.send_message_to_open_ephys('StopRecord')
            sleep(0.1)
        print('Stopping Open Ephys GUI Recording Successful')

        self.compile_recording_data()

        # Stop cumulative plot
        if self.general_settings['Tracking']:
            if hasattr(self, 'PosPlot'):
                print('Stopping Position Plot...')
                self.position_plot.close()
                print('Stopping Position Plot Successful')

        # Close connection with OpenEphysGUI
        self.open_ephys_messenger.close()

        self.recording_closing = False

    def close(self):
        if self.recording_initialized:
            self.cancel_init_devices()
        elif self.recording_active:
            self.stop_rec()
        elif self.recording_closing:
            while self.recording_closing:
                sleep(1)
        else:
            pass


def list_window(items):
    """
    Creates a new window with a list of string items to select from.
    Double clicking returns the string on the item selected

    Input:  list of strings
    Output: selected list element
    """
    dialog_widget = QtWidgets.QDialog()
    dialog_widget.setWindowTitle('Choose option:')
    dialog_widget.resize(200, 200)
    list_widget = QtWidgets.QListWidget()
    list_widget.addItems(items)
    list_widget.itemDoubleClicked.connect(dialog_widget.accept)
    scroll = QtWidgets.QScrollArea()
    scroll.setWidget(list_widget)
    scroll.setWidgetResizable(True)
    dialog_layout = QtWidgets.QVBoxLayout(dialog_widget)
    dialog_layout.addWidget(scroll)
    dialog_widget.show()
    if dialog_widget.exec_():
        selected_item = items[list_widget.currentRow()]
    else:
        selected_item = None

    return selected_item


def add_boolean_setting_to_widget(label, state, value_change_callback, parent, layout):
    checkbox_widget = QtWidgets.QCheckBox(label, parent)
    layout.addWidget(checkbox_widget)
    checkbox_widget.setChecked(state)
    checkbox_widget.stateChanged.connect(lambda: value_change_callback(checkbox_widget.isChecked()))


def add_x_y_setting_to_widget(label, x_val, y_val, value_change_callback, parent, layout):

    widget = QtWidgets.QWidget(parent)
    layout.addWidget(widget)

    label = QtWidgets.QLabel(label, widget)

    xy_widget = QtWidgets.QWidget(widget)

    x_label = QtWidgets.QLabel('    X', xy_widget)
    x_textbox = QtWidgets.QTextEdit(str(x_val), xy_widget)
    y_label = QtWidgets.QLabel('    Y', xy_widget)
    y_textbox = QtWidgets.QTextEdit(str(y_val), xy_widget)
    xy_layout = QtWidgets.QHBoxLayout(xy_widget)
    xy_layout.setContentsMargins(2, 2, 2, 2)
    xy_layout.setSpacing(4)
    xy_layout.addWidget(x_label)
    xy_layout.addWidget(x_textbox)
    xy_layout.addWidget(y_label)
    xy_layout.addWidget(y_textbox)
    x_textbox.setFixedHeight(27)
    y_textbox.setFixedHeight(27)

    x_textbox.textChanged.connect(lambda: value_change_callback([x_textbox.toPlainText(),
                                                                 y_textbox.toPlainText()]))

    widget_layout = QtWidgets.QHBoxLayout(widget)
    widget_layout.setContentsMargins(2, 2, 2, 2)
    widget_layout.setSpacing(2)
    widget_layout.addWidget(label)
    widget_layout.addWidget(xy_widget)

    widget_layout.setStretch(0, 1)
    widget_layout.setStretch(1, 2)


def add_text_setting_to_widget(label, text, value_change_callback, parent, layout, button_kwargs=None):

    widget = QtWidgets.QWidget(parent)
    layout.addWidget(widget)

    label = QtWidgets.QLabel(label, widget)
    textbox = QtWidgets.QTextEdit(text, widget)
    textbox.textChanged.connect(lambda: value_change_callback(textbox.toPlainText()))

    button = None
    if not (button_kwargs is None):
        button = QtWidgets.QPushButton(button_kwargs['label'], widget)
        button.setFixedSize(60, 27)
        button.clicked.connect(lambda: button_kwargs['connect'](textbox))

    textbox.setFixedHeight(27)
    widget_layout = QtWidgets.QHBoxLayout(widget)
    widget_layout.setContentsMargins(2, 2, 2, 2)
    widget_layout.setSpacing(2)
    widget_layout.addWidget(label)
    widget_layout.addWidget(textbox)
    if not (button_kwargs is None):
        widget_layout.addWidget(button)

    if not (button_kwargs is None):
        widget_layout.setStretch(0, 2)
        widget_layout.setStretch(1, 3)
        widget_layout.setStretch(2, 1)
    else:
        widget_layout.setStretch(0, 1)
        widget_layout.setStretch(1, 2)


class ItemDictionaryWidget(QtWidgets.QFrame):

    def __init__(self, label, col_1_label, col_2_label, item_dict, value_change_callback, parent):
        super().__init__(parent=parent)

        self.value_change_callback = value_change_callback

        QtWidgets.QHBoxLayout(self)

        left_panel_widget = QtWidgets.QWidget(self)
        self.layout().addWidget(left_panel_widget)
        left_panel_layout = QtWidgets.QVBoxLayout(left_panel_widget)
        left_panel_layout.setContentsMargins(0, 0, 0, 0)
        left_panel_layout.setSpacing(2)
        left_panel_layout.setAlignment(QtCore.Qt.AlignTop)

        label = QtWidgets.QLabel(label, left_panel_widget)
        left_panel_layout.addWidget(label)
        new_item_button = QtWidgets.QPushButton('New Group', left_panel_widget)
        new_item_button.clicked.connect(lambda: self.create_new_item())
        new_item_button.setFont(QtGui.QFont(new_item_button.fontInfo().family(), 8))
        left_panel_layout.addWidget(new_item_button)

        right_panel_widget = QtWidgets.QWidget(self)
        self.layout().addWidget(right_panel_widget)
        right_panel_layout = QtWidgets.QVBoxLayout(right_panel_widget)
        right_panel_layout.setContentsMargins(0, 0, 0, 0)
        right_panel_layout.setSpacing(2)

        right_panel_header_widget = QtWidgets.QWidget(right_panel_widget)
        right_panel_layout.addWidget(right_panel_header_widget)
        col_1_label = QtWidgets.QLabel(col_1_label, right_panel_header_widget)
        col_2_label = QtWidgets.QLabel(col_2_label, right_panel_header_widget)
        right_panel_header_layout = QtWidgets.QHBoxLayout(right_panel_header_widget)
        right_panel_header_layout.setContentsMargins(0, 0, 0, 0)
        right_panel_header_layout.setSpacing(0)
        right_panel_header_layout.addWidget(col_1_label)
        right_panel_header_layout.addWidget(col_2_label)

        scroll = QtWidgets.QScrollArea(right_panel_header_widget)
        scroll.setWidgetResizable(True)
        scroll.setFixedHeight(70)
        right_panel_layout.addWidget(scroll)
        self.item_widget_list_widget = QtWidgets.QWidget(scroll)
        scroll.setWidget(self.item_widget_list_widget)
        self.item_widget_list_layout = QtWidgets.QVBoxLayout(self.item_widget_list_widget)
        self.item_widget_list_layout.setContentsMargins(0, 0, 0, 0)
        self.item_widget_list_layout.setSpacing(0)

        self.layout().setStretch(0, 1)
        self.layout().setStretch(1, 2)

        self.items = {}

        for key, value in item_dict.items():
            self.create_new_item(col_1_value=key, col_2_value=value)

    def get_item_dict(self):
        return {
            str(item['col_1_textbox'].toPlainText()): {
                'string': str(item['col_2_textbox'].toPlainText()),
                'item_id': item_id,
                'readout_successful': lambda item_id: self.set_item_color(item_id, (100, 255, 100)),
                'readout_failed': lambda item_id: self.set_item_color(item_id, (255, 100, 100))
            }
            for item_id, item in self.items.items()
        }

    def value_changed_event(self):
        return self.value_change_callback(self.get_item_dict())

    def create_new_item(self, col_1_value='', col_2_value=''):

        item_id = uuid4()

        item_widget = QtWidgets.QWidget(self.item_widget_list_widget)
        self.item_widget_list_layout.addWidget(item_widget)
        item_layout = QtWidgets.QHBoxLayout(item_widget)
        item_layout.setContentsMargins(0, 0, 0, 0)
        item_layout.setSpacing(0)

        col_1_textbox = QtWidgets.QTextEdit(col_1_value, item_widget)
        col_1_textbox.textChanged.connect(self.value_changed_event)
        col_1_textbox.setFixedHeight(27)
        item_layout.addWidget(col_1_textbox)

        col_2_textbox = QtWidgets.QTextEdit(col_2_value, item_widget)
        col_2_textbox.textChanged.connect(self.value_changed_event)
        col_2_textbox.setFixedHeight(27)
        item_layout.addWidget(col_2_textbox)

        remove_button = QtWidgets.QPushButton('X', item_widget)
        remove_button.setFixedSize(15, 27)
        item_layout.addWidget(remove_button)
        remove_button.clicked.connect(lambda: self.remove_item(item_id))

        self.items[item_id] = {'col_1_textbox': col_1_textbox,
                               'col_2_textbox': col_2_textbox,
                               'item_widget': item_widget}

        self.value_change_callback(self.get_item_dict())

    def remove_item(self, item_id):
        item = self.items.pop(item_id)
        item['item_widget'].deleteLater()
        self.value_change_callback(self.get_item_dict())

    def set_item_color(self, item_id, color):
        for column_textbox_name in ('col_1_textbox', 'col_2_textbox'):
            p = self.items[item_id][column_textbox_name].viewport().palette()
            p.setColor(self.items[item_id][column_textbox_name].viewport().backgroundRole(), QtGui.QColor(*color))
            self.items[item_id][column_textbox_name].viewport().setPalette(p)


def convert_item_list_string_values_to_channel_map(item_dict):
    channel_map = {}
    for channel_group, item in item_dict.items():
        try:
            if not ('-' in item['string']):
                raise ValueError('There is no "-" in the string')
            first_channel = int(item['string'][:item['string'].find('-')]) - 1
            second_channel = int(item['string'][item['string'].find('-') + 1:])
            if first_channel >= second_channel:
                raise ValueError('Second channel should be a larger value')
            channel_map[channel_group] = {
                'string': item['string'],
                'list': np.arange(first_channel, second_channel, dtype=np.int64)
            }
            item['readout_successful'](item['item_id'])
        except:
            item['readout_failed'](item['item_id'])

    return channel_map


class QOneShotThread(QtCore.QThread):

    def __init__(self, target):
        self._target = target
        super().__init__()

    def run(self) -> None:
        self._target()


class RecordingManagerGUI(QtWidgets.QMainWindow):

    exit_code_to_reboot = -123

    geometry = None

    test_devices_thread = None
    initialize_devices_thread = None
    start_recording_thread = None
    stop_recording_thread = None

    def __init__(self, recording_manager, geometry=None):
        """
        :param RecordingManager recording_manager:
        """
        super().__init__()

        self.recording_manager = recording_manager

        self.setFixedSize(350, 600)

        if not (geometry is None):
            self.restoreGeometry(geometry)

        # Create general GUI layout

        self.central_widget = QtWidgets.QWidget(self)
        self.button_widget = QtWidgets.QWidget(self.central_widget)
        self.button_widget.setMaximumHeight(220)
        self.general_settings_widget = QtWidgets.QWidget(self.central_widget)
        self.current_recording_file_path = QtWidgets.QLabel('Current recording folder:', self.central_widget)
        self.current_recording_file_path.setFixedHeight(40)
        self.general_settings_label = QtWidgets.QLabel('General Settings', self.central_widget)
        self.general_settings_label.setFont(QtGui.QFont('SansSerif', 15))
        self.central_layout = QtWidgets.QVBoxLayout(self.central_widget)
        self.central_layout.addWidget(self.button_widget)
        self.central_layout.addWidget(self.current_recording_file_path)
        self.central_layout.addWidget(self.general_settings_label)
        self.central_layout.addWidget(self.general_settings_widget)
        self.setCentralWidget(self.central_widget)

        # Add buttons to button widget

        # Define buttons
        self.load_last_settings_button = QtWidgets.QPushButton('Load Last Settings', self.button_widget)
        self.load_last_settings_button.clicked.connect(self.load_last_settings_button_callback)
        self.load_settings_button = QtWidgets.QPushButton('Load Settings', self.button_widget)
        self.load_settings_button.clicked.connect(self.load_settings_button_callback)
        self.save_settings_button = QtWidgets.QPushButton('Save Settings', self.button_widget)
        self.save_settings_button.clicked.connect(self.save_settings_button_callback)
        self.camera_settings_button = QtWidgets.QPushButton('Camera Settings', self.button_widget)
        self.camera_settings_button.clicked.connect(self.camera_settings_button_callback)
        self.task_settings_button = QtWidgets.QPushButton('Task Settings', self.button_widget)
        self.task_settings_button.clicked.connect(self.task_settings_button_callback)
        self.test_devices_button = QtWidgets.QPushButton('Test Devices', self.button_widget)
        self.test_devices_button.clicked.connect(self.test_devices_button_callback)
        self.initialize_devices_button = QtWidgets.QPushButton('Initialize Devices', self.button_widget)
        self.initialize_devices_button.clicked.connect(self.initialize_devices_button_callback)
        self.start_recording_button = QtWidgets.QPushButton('Start Recording', self.button_widget)
        self.start_recording_button.clicked.connect(self.start_recording_button_callback)
        self.start_recording_button.setEnabled(False)
        self.stop_recording_button = QtWidgets.QPushButton('Stop Recording', self.button_widget)
        self.stop_recording_button.clicked.connect(self.stop_recording_button_callback)
        self.stop_recording_button.setEnabled(False)

        self.default_button_style = self.stop_recording_button.styleSheet()

        self.stop_recording_button.clicked.connect(lambda: print({'general': self.recording_manager.general_settings,
                                                                  'camera': self.recording_manager.camera_settings,
                                                                  'task': self.recording_manager.task_settings}))

        # Position buttons
        self.button_layout = QtWidgets.QGridLayout(self.button_widget)
        self.button_layout.addWidget(self.load_last_settings_button, 0, 0)
        self.button_layout.addWidget(self.load_settings_button, 1, 0)
        self.button_layout.addWidget(self.save_settings_button, 2, 0)
        self.button_layout.addWidget(self.camera_settings_button, 4, 0)
        self.button_layout.addWidget(self.task_settings_button, 5, 0)
        self.button_layout.addWidget(self.test_devices_button, 0, 1)
        self.button_layout.addWidget(self.initialize_devices_button, 1, 1)
        self.button_layout.addWidget(self.start_recording_button, 2, 1)
        self.button_layout.addWidget(self.stop_recording_button, 3, 1)
        self.button_layout.setHorizontalSpacing(40)
        self.button_layout.setVerticalSpacing(2)

        # Add settings to general settings widget
        self.general_settings_layout = QtWidgets.QVBoxLayout(self.general_settings_widget)
        self.general_settings_layout.setContentsMargins(2, 2, 2, 2)
        self.general_settings_layout.setSpacing(2)

        checkboxes_widget = QtWidgets.QWidget(self.general_settings_widget)
        checkboxes_layout = QtWidgets.QHBoxLayout(checkboxes_widget)
        checkboxes_layout.setContentsMargins(2, 2, 2, 2)
        checkboxes_layout.setSpacing(2)
        self.general_settings_layout.addWidget(checkboxes_widget)
        add_boolean_setting_to_widget(
            'Tracking', self.recording_manager.general_settings['Tracking'],
            lambda state: self.recording_manager.update_general_settings('Tracking', state),
            checkboxes_widget, checkboxes_layout
        )
        add_boolean_setting_to_widget(
            'Task active', self.recording_manager.general_settings['TaskActive'],
            lambda state: self.recording_manager.update_general_settings('TaskActive', state),
            checkboxes_widget, checkboxes_layout
        )

        add_x_y_setting_to_widget(
            'Arena size',
            self.recording_manager.general_settings['arena_size'][0],
            self.recording_manager.general_settings['arena_size'][1],
            lambda text: self.recording_manager.update_general_settings('arena_size', text),
            self.general_settings_widget, self.general_settings_layout
        )
        add_text_setting_to_widget(
            'Experimenter', self.recording_manager.general_settings['experimenter'],
            lambda text: self.recording_manager.update_general_settings('experimenter', text),
            self.general_settings_widget, self.general_settings_layout,
            button_kwargs={'label': 'options',
                           'connect': lambda text_box: self.pick_setting_from_options('experimenter',
                                                                                      text_box)}
        )
        add_text_setting_to_widget(
            'Experiment ID', self.recording_manager.general_settings['experiment_id'],
            lambda text: self.recording_manager.update_general_settings('experiment_id', text),
            self.general_settings_widget, self.general_settings_layout,
            button_kwargs={'label': 'options',
                           'connect': lambda text_box: self.pick_setting_from_options('experiment_id',
                                                                                      text_box)}
        )
        add_text_setting_to_widget(
            'Animal ID', self.recording_manager.general_settings['animal'],
            lambda text: self.recording_manager.update_general_settings('animal', text),
            self.general_settings_widget, self.general_settings_layout,
            button_kwargs={'label': 'options',
                           'connect': lambda text_box: self.pick_setting_from_options('animal',
                                                                                      text_box)}
        )
        add_text_setting_to_widget(
            'Bad channel list', self.recording_manager.general_settings['badChan'],
            lambda text: self.recording_manager.update_general_settings('badChan', text),
            self.general_settings_widget, self.general_settings_layout,
            button_kwargs={'label': 'options',
                           'connect': lambda text_box: self.pick_setting_from_options('badChan',
                                                                                      text_box)}
        )

        channel_map_widget = ItemDictionaryWidget(
            'Channel map', 'Group', 'Channel range',
            {key: item['string']
             for key, item in self.recording_manager.general_settings['channel_map'].items()},
            lambda item_dict: self.recording_manager.update_general_settings(
                'channel_map',
                convert_item_list_string_values_to_channel_map(item_dict)
            ),
            self.general_settings_widget
        )
        self.general_settings_layout.addWidget(channel_map_widget)

    def pick_setting_from_options(self, setting_key, setting_text_box):
        setting_options = self.recording_manager.list_general_setting_options(setting_key)
        selected_item = list_window(setting_options)
        if not (selected_item is None):
            setting_text_box.setPlainText(selected_item)

    def save_settings_button_callback(self):
        fpath = HFunc.openSingleFileDialog('save', suffix='nwb', caption='Save file name and location')
        if not (fpath is None):
            self.recording_manager.save_settings(fpath)

    def load_settings_button_callback(self):
        fpath = HFunc.openSingleFileDialog('load', suffix='nwb', caption='Select file to load')
        if not (fpath is None):
            self.recording_manager.load_settings(fpath)
            self.restart()

    def load_last_settings_button_callback(self):
        try:
            self.recording_manager.load_last_settings()
            self.restart()
        except ValueError as e:
            HFunc.show_message(str(e))

    def camera_settings_button_callback(self):
        camera_settings_app = self.recording_manager.camera_settings_app()
        CameraSettingsGUI(camera_settings_app, parent=self)

    def task_settings_button_callback(self):
        task_settings_app = self.recording_manager.task_settings_app()
        TaskSettingsGUI(task_settings_app, parent=self)

    @staticmethod
    def display_device_status_message(device_status):

        n_unavailable = sum([not device['available'] for device in device_status])

        if n_unavailable == 0:
            message = 'All devices available.'
        else:
            message = '{} of {} devices unavailable!'.format(n_unavailable, len(device_status))

        message_more = ''
        for device in device_status:
            if device['available']:
                message_more += '{} online\n'.format(device['name'])
            else:
                message_more += '{} OFFLINE\n'.format(device['name'])

        HFunc.show_message(message, message_more=message_more)

    def test_devices(self):
        try:
            device_status = self.recording_manager.test_devices()
            self.display_device_status_message(device_status)
        except RecordingManagerException as e:
            HFunc.show_message(str(e))

    def test_devices_finished(self):
        self.test_devices_button.setStyleSheet(self.default_button_style)
        while self.test_devices_thread.isRunning():
            sleep(0.01)
        self.test_devices_thread = None

    def test_devices_button_callback(self):
        self.test_devices_button.setStyleSheet('background-color: red')
        self.test_devices_thread = QOneShotThread(self.test_devices)
        self.test_devices_thread.finished.connect(self.test_devices_finished)
        self.test_devices_thread.start()

    def initialize_devices(self):
        try:
            self.recording_manager.init_devices()
        except RecordingManagerException as e:
            print(e)
            HFunc.show_message(str(e))

    def initialize_devices_finished(self):
        if self.recording_manager.recording_initialized:
            self.initialize_devices_button.setStyleSheet('background-color: green')
            self.initialize_devices_button.setEnabled(False)
            self.start_recording_button.setEnabled(True)
        else:
            self.initialize_devices_button.setStyleSheet(self.default_button_style)
        while self.initialize_devices_thread.isRunning():
            sleep(0.01)
        self.initialize_devices_thread = None

    def initialize_devices_button_callback(self):
        self.initialize_devices_button.setStyleSheet('background-color: red')
        self.initialize_devices_thread = QOneShotThread(self.initialize_devices)
        self.initialize_devices_thread.finished.connect(self.initialize_devices_finished)
        self.initialize_devices_thread.start()

    def start_recording(self):
        try:
            self.recording_manager.start_rec()
        except RecordingManagerException as e:
            print(e)
            HFunc.show_message(str(e))

    def start_recording_finished(self):
        if self.recording_manager.recording_active:
            self.start_recording_button.setStyleSheet('background-color: green')
            self.start_recording_button.setEnabled(False)
            self.stop_recording_button.setEnabled(True)
        else:
            self.start_recording_button.setStyleSheet(self.default_button_style)
        while self.start_recording_thread.isRunning():
            sleep(0.01)
        self.start_recording_thread = None

    def start_recording_button_callback(self):
        self.start_recording_button.setStyleSheet('background-color: red')
        self.start_recording_thread = QOneShotThread(self.start_recording)
        self.start_recording_thread.finished.connect(self.start_recording_finished)
        self.start_recording_thread.start()

    def stop_recording(self):
        try:
            self.recording_manager.stop_rec()
        except RecordingManagerException as e:
            print(e)
            HFunc.show_message(str(e))

    def stop_recording_finished(self):
        if not self.recording_manager.recording_active:
            self.stop_recording_button.setStyleSheet(self.default_button_style)
            self.stop_recording_button.setEnabled(False)
            self.initialize_devices_button.setEnabled(True)
            HFunc.show_message('Recording Completed Successfully.')
        else:
            self.stop_recording_button.setStyleSheet(self.default_button_style)
        while self.stop_recording_thread.isRunning():
            sleep(0.01)
        self.stop_recording_thread = None

    def stop_recording_button_callback(self):
        self.stop_recording_button.setStyleSheet('background-color: red')
        self.start_recording_button.setStyleSheet(self.default_button_style)
        self.stop_recording_thread = QOneShotThread(self.stop_recording)
        self.stop_recording_thread.finished.connect(self.stop_recording_finished)
        self.stop_recording_thread.start()

    def restart(self):
        self.geometry = self.saveGeometry()
        QtCore.QCoreApplication.exit(self.exit_code_to_reboot)


def recording_manager_gui_with_load_settings_loop(recording_manager):

    # While loop allows restarting GUI without restarting the RecordingManager application
    exit_code = RecordingManagerGUI.exit_code_to_reboot
    geometry = None
    while exit_code == RecordingManagerGUI.exit_code_to_reboot:

        # Start application
        app = QtWidgets.QApplication(sys.argv)
        window = RecordingManagerGUI(recording_manager, geometry=geometry)
        window.show()
        exit_code = app.exec_()
        geometry = deepcopy(window.geometry)
        app.exit(exit_code)
        del app


def main():
    recording_manager = RecordingManager()
    recording_manager_gui_with_load_settings_loop(recording_manager)
    recording_manager.close()


if __name__ == '__main__':
    main()




import os
from copy import deepcopy
from importlib import import_module
from PyQt5 import QtWidgets

from openEPhys_DACQ import NWBio
from openEPhys_DACQ.HelperFunctions import openSingleFileDialog, clearLayout


class TaskSettingsApp(object):

    def __init__(self, arena_size, apply_settings_method, task_settings=None):

        self._arena_size = arena_size
        self.apply_settings_method = apply_settings_method
        self._task_settings = {} if task_settings is None else task_settings
        self._task_settings['arena_size'] = arena_size

    @property
    def arena_size(self):
        return deepcopy(self._arena_size)

    @property
    def task_settings(self):
        return deepcopy(self._task_settings)

    def update_settings(self, key, value):
        self._task_settings[key] = value

    def load_settings(self, fpath):
        self._task_settings = NWBio.load_settings(fpath, path='/TaskSettings/')

    def save_settings(self, fpath):
        NWBio.save_settings(fpath, self.task_settings, path='/TaskSettings/')

    def apply(self):
        self.apply_settings_method(self.task_settings)


class TaskSettingsGUI(QtWidgets.QDialog):

    def __init__(self, task_settings_app, parent):
        """
        :param parent:
        :param TaskSettingsApp task_settings_app:
        """
        super().__init__(parent=parent)

        self.task_settings_app = task_settings_app

        self.task_specific_gui = None

        self.resize(400, 150)
        self.setWindowTitle('Task Settings')

        # Create top menu items
        self.task_selection_list_widget = QtWidgets.QListWidget(self)
        self.task_selection_list_widget.setMaximumWidth(300)
        self.task_selection_list_widget.itemSelectionChanged.connect(self.task_selection_list_callback)
        load_button = QtWidgets.QPushButton('Load')
        load_button.setMaximumWidth(100)
        load_button.clicked.connect(self.load_button_callback)
        save_button = QtWidgets.QPushButton('Save')
        save_button.setMaximumWidth(100)
        save_button.clicked.connect(self.save_button_callback)
        apply_button = QtWidgets.QPushButton('Apply')
        apply_button.setMaximumWidth(100)
        apply_button.clicked.connect(self.apply_button_callback)
        cancel_button = QtWidgets.QPushButton('Cancel')
        cancel_button.setMaximumWidth(100)
        cancel_button.clicked.connect(self.cancel_button_callback)

        self.setLayout(QtWidgets.QHBoxLayout(self))

        # Create top menu and general settings widget
        top_menu_general_settings_widget = QtWidgets.QWidget(self)
        top_menu_general_settings_layout = QtWidgets.QVBoxLayout(top_menu_general_settings_widget)
        self.layout().addWidget(top_menu_general_settings_widget)

        # Create top menu layout
        top_menu_widget = QtWidgets.QWidget(top_menu_general_settings_widget)
        top_menu_widget.setFixedSize(300, 125)
        top_menu_general_settings_layout.addWidget(top_menu_widget)
        top_menu_layout = QtWidgets.QHBoxLayout(top_menu_widget)
        top_menu_layout.addWidget(self.task_selection_list_widget)
        top_menu_buttons_widget = QtWidgets.QWidget(top_menu_widget)
        top_menu_layout.addWidget(top_menu_buttons_widget)
        top_menu_buttons_layout = QtWidgets.QVBoxLayout(top_menu_buttons_widget)
        top_menu_buttons_layout.addWidget(load_button)
        top_menu_buttons_layout.addWidget(save_button)
        top_menu_buttons_layout.addWidget(apply_button)
        top_menu_buttons_layout.addWidget(cancel_button)
        top_menu_buttons_layout.setContentsMargins(0, 0, 0, 0)
        top_menu_buttons_layout.setSpacing(2)

        # Create task general menu items
        self.general_settings_widget = QtWidgets.QWidget(top_menu_general_settings_widget)
        self.general_settings_layout = QtWidgets.QVBoxLayout(self.general_settings_widget)
        top_menu_general_settings_layout.addWidget(self.general_settings_widget)

        # Create task specific menu items
        self.further_settings_widget = QtWidgets.QWidget(self)
        self.further_settings_layout = QtWidgets.QHBoxLayout(self.further_settings_widget)
        self.layout().addWidget(self.further_settings_widget)

        # Populate task selection list
        self.task_selection_list_items = {}
        tasks_dir = os.path.join(os.path.dirname(NWBio.__file__), 'Tasks')
        for task_name in os.listdir(tasks_dir):
            if os.path.isdir(os.path.join(tasks_dir, task_name)) and not task_name.startswith('__'):
                self.task_selection_list_items[task_name] = \
                    QtWidgets.QListWidgetItem(task_name, self.task_selection_list_widget)
                self.task_selection_list_widget.addItem(self.task_selection_list_items[task_name])

        # Load task settings if any available on task_settings_app
        if 'name' in self.task_settings_app.task_settings:
            self.select_task_selection_list_item(self.task_settings_app.task_settings['name'])
            self.update_task_specific_settings_data()

        self.exec()

    def select_task_selection_list_item(self, task_name):
        self.task_selection_list_widget.setCurrentItem(self.task_selection_list_items[task_name])

    def task_selection_list_callback(self, item=None):
        if item is None:
            item = self.task_selection_list_widget.currentItem()
        current_task = str(item.text())
        self.load_task_specific_gui(current_task)

    def load_task_specific_gui(self, current_task):
        # Load the GUI for currently selected task
        task_module = import_module('Tasks.' + current_task + '.Task')
        clearLayout(self.general_settings_layout)
        clearLayout(self.further_settings_layout)
        self.task_specific_gui = task_module.SettingsGUI(self.general_settings_layout,
                                                         self.further_settings_layout,
                                                         self.task_settings_app.arena_size)
        self.resize(self.task_specific_gui.min_size[0], self.task_specific_gui.min_size[1])

    def load_button_callback(self):
        fpath = openSingleFileDialog('load', suffix='nwb', caption='Select file to load')
        self.task_settings_app.load_settings(fpath)
        self.select_task_selection_list_item(self.task_settings_app.task_settings['name'])
        self.update_task_specific_settings_data()

    def update_task_settings_app_data(self):
        task_settings = self.task_specific_gui.export_settings_from_gui()
        task_settings['name'] = str(self.task_selection_list_widget.currentItem().text())
        for key, value in task_settings.items():
            self.task_settings_app.update_settings(key, value)

    def update_task_specific_settings_data(self):
        task_settings = deepcopy(self.task_settings_app.task_settings)
        _ = task_settings.pop('name')
        self.task_specific_gui.import_settings_to_gui(task_settings)

    def save_button_callback(self):
        self.update_task_settings_app_data()
        fpath = openSingleFileDialog('save', suffix='nwb', caption='Save file name and location')
        self.task_settings_app.save_settings(fpath)

    def apply_button_callback(self):
        self.update_task_settings_app_data()
        self.task_settings_app.apply()
        self.close()

    def cancel_button_callback(self):
        self.close()
